/*****************************************************************************/
#ifdef COMMENTS_WITH_COMMENTS
/*
                                  compose.c

All things involved in the composition and sending of a mail message.

When sending a message, if there are soyMAIL attachment files that have been
checked for inclusion a "multipart/mixed" MIME message is generated.  Each
attachment with a "text/.." content-type is quoted-printable encoded in a
separate part.  Attachments with non-text content are provided as base-64
encoded parts.  If there is a message body this is appropriately encoded and
supplied as the first part with a content-type of "text/plain".

If there are no attachments files present or checked the message body is
appropriately encoded and made into an in-line MIME message with a plain-text
content-type.

If any one of the address fields (To, CC or BCC) contain a non-VMS address the
SMTP Message Transport Agent is used to send the message.  If the address
fields only contain what looks like VMS Mail addresses the message is sent via
VMS callable Mail.


HTML EDITING
------------
This rather neat capability is provided using stand-along, JavaScript-driven,
HTML text editors that can be attached to the composition text area.  soyMAIL
has some hooks that recognise and support such editors, and when used to
compose a message the content is made into quoted-printable encoded HTML for
transmission.  They are made available using the configuration directives

  [html-editor-load]     JavaScript code to load an initialise the editor

So far it has been tried successfully with

  o  FCKeditor v2.2      http://www.fckeditor.net/
                         http://sourceforge.net/projects/fckeditor/

  o  TinyMCE v2.0        http://tinymce.moxiecode.com/
     MoxieCode Systems   http://www.moxiecode.com/

Make your choice.  Install it independently of soyMAIL.  Check configuration
requirements for the editor.  Modify the soyMAIL configuration directives to
provide the path to the editor and the initialization JavaScript code, both of
which will be provided in the editor installation guidelines, and voila!  The
function htmlEditorLoad() must be present (even if empty as in the case of
TinyMCE), being an onload= target for the document.


EXAMPLE FOR TINYMCE
-------------------
A functional configuration for TnyMCE.

  [html-editor-load]
  <script type="text/javascript" 
          src="/tinymce/jscripts/tiny_mce/tiny_mce.js"></script>
  <script type="text/javascript">

  tinyMCE.init({
     mode : "exact",
     elements : "compose_text",
     theme : "advanced",
     plugins: "preview,print,insertdatetime",
     theme_advanced_buttons1_add: "fontselect,fontsizeselect",
     theme_advanced_buttons2_add: "preview,print",
     theme_advanced_buttons3_add: "insertdate,inserttime",
     theme_advanced_buttons3_add_before: "forecolor,backcolor"
  });

  function htmlEditorLoad() { /* nothing required for TinyMCE! */ }
  </script>

This of course requires a mapping rule to provide access to the TinyMCE
installation.

  pass /tinymce/* /$1$disk/tinymce/*


EXAMPLE FOR FCKEDITOR
---------------------

  [html-editor-load]
  <script type="text/javascript" src="/FCKeditor/fckeditor.js"></script>
  <script type="text/javascript">
  function htmlEditorLoad()
  {
     var oFCKeditor = new FCKeditor( 'compose_text' ) ;
     oFCKeditor.BasePath = "/fckeditor/" ;
     oFCKeditor.ReplaceTextarea() ;
  }
  </script>

And a similar mapping requirement exists for FCKeditor.

  pass /fckeditor/* /$1$disk/fckeditor/*



COPYRIGHT
---------
Copyright (C) 2005-2007 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY
---------------
22-MAR-2007  MGD  ComposeBodyTextMassage() refine eol space handling
12-FEB-2007  MGD  draft maintenance now performed by DRAFT.C module
21-OCT-2006  MGD  ComposeSendMessage() if 'delete all attachments' checkbox 
                    delete all attachments after successful send
20-JUL-2006  MGD  rework reply-all address compilation
                  try to make font width allowances for monospace
19-JUL-2006  MGD  bugfix; ComposeSendMessage() check reply-to for VMS address
                  bugfix; ComposePage() HTML-escape the hidden signature
                  bugfix; ComposeTextMassage() retain quote-char as a string
06-JUL-2006  MGD  ComposeCopySelf() to refine interactions between VMS mail
                    addresses, [smtp-default-host], and copy-self, etc.
01-JUL-2006  MGD  add [forward]ing (from message read page) of attachments
                    with ComposeReplyAttachList(), ComposeForwardAttachList()
                    and changes to ComposeMime() to add the forwarded parts
24-JUN-2006  MGD  significant enhancements to JavaScript functionality
15-JUN-2006  MGD  ComposePage() improve layout and resizing of addressing
12-JUN-2006  MGD  ComposeRfcDate() to derive RFC822 "Date:" from client
                    to better represent message origination date/time
25-APR-2006  MGD  reply-to/all to reply-to: if available otherwise from:
                  ComposePage() do not infinitely increase the number of
                    'RE:' or 'FW:' in subject lines
14-APR-2006  MGD  bugfix; ComposeSelfAddress() size for calloc()
04-APR-2006  MGD  add [send-close] and ComposeSendClose()
15-MAR-2006  MGD  use form field TABINDEX attribute to control focus
13-MAR-2006  MGD  bugfix; ComposeSelfAddress() 'len' initialization
11-MAR-2006  MGD  move input focus to the field most appropriate to the stage
                    of message composition, defaulting to the message text
                  bugfix; ComposeSelfAddress() with rationalization of 'self' 
                    address generation and using this for the "From:" header
25-FEB-2006  MGD  ComposePageRequest() support VMS copy-send/forward/reply
                  ComposeSendMessage() expand MX_.. logical names that
                    can be used to generate the "From:" address
                  ComposePage() [signature] button only if the user has a
                    signature file specified
14-FEB-2006  MGD  when used from the message page the main menu [compose]
                    button should provide a blank message, when used from the
                    compose page it should propagate any in-composition message
09-FEB-2006  MGD  bugfix; default charset should be ISO-8859-1 not IS0-8859-1
                    (actually went to using the macro SOY_DEFAULT_CHARSET)
06-FEB-2006  MGD  modified [reply-to] option to [smtp-from] and in conjunction
                    with [personal-name] have ComposePageRequest() generate a
                    "From:" header more in line with TCPIP$SMTP_FROM behaviour
                  add explicit [reply-to] button and field so that messages
                    can individually have a reply-to set against them
                  ComposeMime() add client traceability information to
                    "Received:" header field
01-FEB-2005  MGD  initial
*/
#endif /* COMMENTS_WITH_COMMENTS */
/*****************************************************************************/

#ifdef SOYMAIL_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unixio.h>

/* VMS related header files */
#include <descrip.h>
#include <mailmsgdef.h>
#include <rmsdef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header file */
#include "soymail.h"
#include "address.h"
#include "attach.h"
#include "callmail.h"
#include "cgilib.h"
#include "compose.h"
#include "config.h"
#include "draft.h"
#include "html.h"
#include "mainmenu.h"
#include "mta.h"
#include "options.h"
#include "other.h"
#include "message.h"
#include "request.h"
#include "sendmail.h"

#define FI_LI __FILE__, __LINE__

/* global storage */

/* prototypes */

int sys$gettim (__unknown_params);
int lib$cvt_vectim(__unknown_params);

/* external storage */

extern BOOL  Debug,
             WatchEnabled;

extern int  VmsVersion;

extern char  CurrentVmsTimeString[],
             LocalHostName[],
             SoftwareId[],
             SoftwareVn[];

extern  CONFIG_DATA SoyMailConfig;
extern VMS_MAIL_USER VmsMailUser;

/*****************************************************************************/
/*
Assemble the request data related to this message composition.
*/

BOOL ComposePageRequest (REQUEST_DATA *rdptr)

{
   BOOL  MsgReply, MsgReplyAll,
         MsgForward, MsgEditAsNew;
   char  *cptr, *sptr,
         *CcPtr, *BccPtr, *ReplyToPtr;
   COMPOSE_DATA  *cdptr;
   MIME_DATA  *mdptr;
   USER_OPTIONS  *uoptr;
   REQUEST_DATA  *sdptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("ComposePageRequest()");

   /* a bit of sanity checking */
   if (!rdptr->PrivateAccess) ErrorExit (SS$_BUGCHECK, FI_LI);

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;
   sdptr = (REQUEST_DATA*)rdptr->StateDataPtr;
   uoptr = &rdptr->UserOptions;
   cdptr = &rdptr->ComposeData;

   /* keep track of the message ID */
   vmptr->MessageId = sdptr->MessageId;

   if (SoyMailConfig.VmsOccluded)
   {
      /* no relevant VMS options to be used in message composition */
      muptr->VmsMailCopyForward =
         muptr->VmsMailCopyReply =  
         muptr->VmsMailCopySend = 0;
      muptr->VmsMailPersonalNameLength = 0;
      muptr->VmsMailPersonalName[0] = '\0';
      muptr->VmsMailSigFile[0] = '\0';
      muptr->VmsMailSigFileLength = 0;
   }

   /* this also is used as the "from:" address */
   ComposeSelfAddress (rdptr);

   if (rdptr->PrevPageIdent == PAGE_COMPOSE)
   {
      /*************************/
      /* from the compose page */
      /*************************/

      /* propagate selected items from (any) previous request */
      cdptr->AlreadySent = sdptr->ComposeData.AlreadySent;
      cdptr->ContentTypeIs = sdptr->ComposeData.ContentTypeIs;
      cdptr->CopiedSelf = sdptr->ComposeData.CopiedSelf;
      cdptr->EditorType = sdptr->ComposeData.EditorType;
      cdptr->MsgForward = sdptr->ComposeData.MsgForward;
      cdptr->MsgReply = sdptr->ComposeData.MsgReply;

      CGIVAR (cdptr->ToPtr, "FORM_COMPOSE_TO");
      CGIVARNULL (CcPtr, "FORM_COMPOSE_CC");
      if (!(cdptr->CcPtr = CcPtr)) cdptr->CcPtr = "";
      CGIVARNULL (BccPtr, "FORM_COMPOSE_BCC");
      if (!(cdptr->BccPtr = BccPtr)) cdptr->BccPtr = "";
      CGIVARNULL (ReplyToPtr, "FORM_COMPOSE_REPLY_TO");
      if (!(cdptr->ReplyToPtr = ReplyToPtr)) cdptr->ReplyToPtr = "";
      CGIVAR (cdptr->SubjPtr, "FORM_COMPOSE_SUBJ");
      CGIVAR (cdptr->TextPtr, "FORM_COMPOSE_TEXT");
      CGIVAR (cdptr->CharSetPtr, "FORM_COMPOSE_CHARSET");

      strcpy (cdptr->DraftCreated, sdptr->ComposeData.DraftCreated);
      cdptr->PrevDraftHash = sdptr->ComposeData.DraftHash;
      cdptr->DraftHash = DraftHash (cdptr);

      CGIVARNULL (cptr, "FORM_COMPOSE_EDIT_ROWS");
      if (cptr) cdptr->EditRows = atoi(cptr);
      CGIVARNULL (cptr, "FORM_COMPOSE_EDIT_COLS");
      if (cptr) cdptr->EditCols = atoi(cptr); 
      CGIVARNULL (cptr, "FORM_COMPOSE_WRAP_AT");
      if (cptr) cdptr->WrapAt = atoi(cptr);

      CGIVARNULL (cptr, "FORM_COMPOSE_BTN_ROWS");
      if (cptr)
      {
         while (*cptr && isspace(*cptr)) cptr++;
         if (isdigit(*cptr))
         {
            cdptr->EditRows = atoi(cptr);  
            switch (cdptr->EditRows)
            {
                case 20 : cdptr->EditRows = 30; break;  
                case 30 : cdptr->EditRows = 40; break;  
                case 40 : cdptr->EditRows = 60; break;  
                default : cdptr->EditRows = 20;
            }
         }
      }

      CGIVARNULL (cptr, "FORM_COMPOSE_BTN_COLS");
      if (cptr)
      {
         while (*cptr && isspace(*cptr)) cptr++;
         if (isdigit(*cptr))
         {
            cdptr->EditCols = atoi(cptr);  
            switch (cdptr->EditCols)
            {
                case 72 : cdptr->EditCols = 130; break;  
                default : cdptr->EditCols = 72;
            }
         }
      }

      CGIVARNULL (cptr, "FORM_CONTACT_ADDR");
      if (cptr) while (*cptr && isspace(*cptr)) cptr++;
      if (cptr && *cptr) cdptr->ContactListPtr = cptr;

      CGIVARNULL (cptr, "FORM_COMPOSE_BTN");
      if (cptr)
      {
         if (LangSame ("compose_to", cptr))
         {
            if (cptr = cdptr->ContactListPtr)
            {
               cptr = ComposeCheckForSelf (rdptr, cptr);
               cdptr->ToPtr = AddressListAppend (cdptr->ToPtr, cptr);
               cptr = AddressListMassage (cptr, TRUE);
               StatusMessage (FI_LI, 0, "%s %s: %s",
                              LangFor("to"), LangFor("compose_add"),
                              HTML_ESCAPE(cptr));
            }
            else
               StatusMessage (FI_LI, 1, LangFor("compose_add_none"));
         }
         else
         if (LangSame ("compose_cc", cptr))
         {
            if (cptr = cdptr->ContactListPtr)
            {
               cptr = ComposeCheckForSelf (rdptr, cptr);
               cdptr->CcPtr = AddressListAppend (cdptr->CcPtr, cptr);
               cptr = AddressListMassage (cptr, TRUE);
               StatusMessage (FI_LI, 0, "%s %s: %s",
                              LangFor("cc"), LangFor("compose_add"),
                              HTML_ESCAPE(cptr));
            }
            else
            if (CcPtr && !CcPtr[0])
            {
               cdptr->ShowCcAddr = FALSE;
               StatusMessage (FI_LI, 0, "%s %s",
                              LangFor("compose"), LangFor("page_opened"));
            }
            else
            if (!CcPtr)
            {
               cdptr->ShowCcAddr = TRUE;
               rdptr->FieldFocusPtr = "document.getElementById(\'compose_cc\').focus();";
               StatusMessage (FI_LI, 0, "%s %s",
                              LangFor("compose"), LangFor("page_opened"));
            }
            else
               StatusMessage (FI_LI, 1, LangFor("compose_add_none"));
         }
         else
         if (LangSame ("compose_bcc", cptr))
         {
            if (cptr = cdptr->ContactListPtr)
            {
               cptr = ComposeCheckForSelf (rdptr, cptr);
               if (AddressIsRfc (cptr))
               {
                  cdptr->BccPtr = AddressListAppend (cdptr->BccPtr, cptr);
                  cptr = AddressListMassage (cptr, TRUE);
                  StatusMessage (FI_LI, 0, "%s %s: %s",
                                 LangFor("bcc"), LangFor("compose_add"),
                                 HTML_ESCAPE(cptr));
               }
               else
                  StatusMessage (FI_LI, 1, "%s: %s",
                                 LangFor("compose_bcc_vms_address"),
                                 HTML_ESCAPE(cptr));
            }
            else
            if (BccPtr && !BccPtr[0])
            {
               cdptr->ShowBccAddr = FALSE;
               StatusMessage (FI_LI, 0, "%s %s",
                              LangFor("compose"), LangFor("page_opened"));
            }
            else
            if (!BccPtr)
            {
               cdptr->ShowBccAddr = TRUE;
               rdptr->FieldFocusPtr = "document.getElementById(\'compose_bcc\').focus();";
               StatusMessage (FI_LI, 0, "%s %s",
                              LangFor("compose"), LangFor("page_opened"));
            }
            else
               StatusMessage (FI_LI, 1, LangFor("compose_add_none"));
         }
         else
         if (LangSame ("compose_reply_to", cptr))
         {
            if (cptr = cdptr->ContactListPtr)
            {
               cptr = ComposeCheckForSelf (rdptr, cptr);
               if (AddressIsRfc (cptr))
               {
                  sptr = calloc (1, strlen(cptr)+1);
                  if (!sptr) ErrorExit (vaxc$errno, FI_LI);
                  strcpy (cdptr->ReplyToPtr = sptr, cptr);
                  StatusMessage (FI_LI, 0, "%s %s: %s",
                                 LangFor("compose_reply_to"),
                                 LangFor("compose_add"),
                                 HTML_ESCAPE(cptr));
               }
               else
                  StatusMessage (FI_LI, 1, "%s: %s",
                                 LangFor("compose_reply_to_vms_address"),
                                 HTML_ESCAPE(cptr));
            }
            else
            if (ReplyToPtr && !ReplyToPtr[0])
            {
               cdptr->ShowBccAddr = FALSE;
               StatusMessage (FI_LI, 0, "%s %s",
                              LangFor("compose"), LangFor("page_opened"));
            }
            else
            if (!ReplyToPtr)
            {
               cdptr->ShowReplyToAddr = TRUE;
               rdptr->FieldFocusPtr = "document.getElementById(\'compose_reply_to\').focus();";
               StatusMessage (FI_LI, 0, "%s %s",
                              LangFor("compose"), LangFor("page_opened"));
            }
            else
               StatusMessage (FI_LI, 1, LangFor("compose_add_none"));
         }
         else
         if (LangSame ("compose_send", cptr))
            ComposeSendMessage (rdptr);
         else
         if (LangSame ("compose_send_close", cptr))
         {
            ComposeSendMessage (rdptr);
            if (cdptr->AlreadySent)
            {
               ComposeSendClose (rdptr);
               return (TRUE);
            }
         }
         else
         if (LangSame ("compose_sign", cptr))
         {
            OptionsSigFileLoad (rdptr);
            StatusMessage (FI_LI, 0, LangFor("compose_sign_done"));
         }
         else
         if (LangSame ("compose_save", cptr))
            DraftSave (rdptr);
         else
         if (LangSame ("compose_restore", cptr))
         {
            /* list the "Draft Items" folder */
            strcpy (rdptr->MailFileName, "MAIL");
            rdptr->MailFileNameLength = strlen(rdptr->MailFileName);
            strcpy (muptr->VmsMailFileName, rdptr->MailFileName);
            muptr->VmsMailFileNameLength = rdptr->MailFileNameLength;

            strcpy (rdptr->FolderName, uoptr->FolderDraftItems);
            rdptr->FolderNameLength = strlen(rdptr->FolderName);
            strcpy (muptr->VmsMailFolderName, rdptr->FolderName);
            muptr->VmsMailFolderNameLength = rdptr->FolderNameLength;

            MessageFolderPage (rdptr);
            return (TRUE);
         }
         else
         if (LangSame ("compose_clear", cptr))
         {
            cdptr->TextPtr = "";
            StatusMessage (FI_LI, 0, LangFor("compose_clear_done"));
         }
         else
         if (LangSame ("compose_new", cptr))
         {
            cdptr->FromPtr = cdptr->ToPtr = cdptr->CcPtr =
               cdptr->BccPtr = cdptr->ReplyToPtr = cdptr->SubjPtr =
               cdptr->TextPtr = cdptr->AttribPtr = NULL;
            cdptr->AlreadySent = cdptr->ContentTypeIs =
               cdptr->MsgForward = cdptr->MsgReply = 0;
            rdptr->FieldFocusPtr = "document.getElementById(\'compose_to\').focus();";
            StatusMessage (FI_LI, 0, LangFor("compose_new_done"));
         }
         else
         if (LangSame ("compose_wrap_off", cptr))
         {
            /* yes this is slightly backward! receipt of 'off' means 'on' */
            cdptr->WrapAt = cdptr->EditCols;
            StatusMessage (FI_LI, 0, "%s.", LangFor("compose_wrap_on"));
         }
         else
         if (LangSame ("compose_wrap_on", cptr))
         {
            /* yes this is slightly backward! receipt of 'on' means 'off' */
            cdptr->WrapAt = 0;
            StatusMessage (FI_LI, 0, "%s.", LangFor("compose_wrap_off"));
         }
         else
         if (LangSame ("compose_edit_html", cptr))
         {
            cdptr->EditorType = COMPOSE_EDITOR_HTML;
            StatusMessage (FI_LI, 0, "%s.", LangFor("compose_edit_html"));
         }
         else
         if (LangSame ("compose_edit_plain", cptr))
         {
            cdptr->EditorType = COMPOSE_EDITOR_PLAIN;
            StatusMessage (FI_LI, 0, "%s.", LangFor("compose_edit_plain"));
         }
         else
         if (LangSame ("open", cptr))
         {
            /* open the contact list */
            StatusMessage (FI_LI, 0, "%s %s",
                           LangFor("compose"), LangFor("page_opened"));
         }
         else
            ErrorExit (SS$_BUGCHECK, FI_LI);
      }
      else
      {
         if (AttachComposeRequest (rdptr)) return (TRUE);
         if (muptr->VmsMailCopySend) ComposeCopySelf (rdptr);
         rdptr->FieldFocusPtr = "document.getElementById(\'compose_to\').focus();";
         StatusMessage (FI_LI, 0, "%s %s",
                        LangFor("compose"), LangFor("page_opened"));
      }
   }
   else
   if (rdptr->PrevPageIdent == PAGE_MESSAGE)
   {
      /**************************/
      /* from message read page */
      /**************************/

      MsgReply = MsgReplyAll = MsgForward = MsgEditAsNew = FALSE;
      CGIVARNULL (cptr, "FORM_MSG_REPLY");
      if (!cptr) CGIVARNULL (cptr, "FORM_MSG_REPLY_POPUP");
      if (cptr) cdptr->MsgReply = MsgReply = TRUE;
      if (!cptr)
      {
         CGIVARNULL (cptr, "FORM_MSG_REPLY_ALL");
         if (!cptr) CGIVARNULL (cptr, "FORM_MSG_REPLY_ALL_POPUP");
         if (cptr) cdptr->MsgReply = MsgReplyAll = TRUE;
      }
      if (!cptr)
      {
         CGIVARNULL (cptr, "FORM_MSG_FORWARD");
         if (!cptr) CGIVARNULL (cptr, "FORM_MSG_FORWARD_POPUP");
         if (cptr) cdptr->MsgForward = MsgForward = TRUE;
         rdptr->FieldFocusPtr = "document.getElementById(\'compose_to\').focus();";
      }
      if (!cptr)
      {
         CGIVARNULL (cptr, "FORM_MSG_EDIT_NEW");
         if (!cptr) CGIVARNULL (cptr, "FORM_MSG_EDIT_NEW_POPUP");
         if (cptr) MsgEditAsNew = TRUE;
         rdptr->FieldFocusPtr = "document.getElementById(\'compose_to\').focus();";
      }

      if (MsgReply || MsgReplyAll || MsgForward || MsgEditAsNew)
      {
         CallMailMessageGet (muptr, NULL);
         if (VMSnok (muptr->MailVmsStatus))
         {
            StatusMessage (FI_LI, 1, "MAIL: %s.",
                           SysGetMsg(muptr->MailVmsStatus));
            ComposePage (rdptr);
            return (TRUE);
         }
         MessageBodyProcess (vmptr);

         /* note that this is using the state (i.e. previous request) data */
         if (SoyMailConfig.HtmlEditorLoad &&
             sdptr->ComposeData.ContentTypeIs == MIME_CONTENT_TEXT_HTML)
            mdptr = MessageBodyIdentify (vmptr, OPTIONS_READ_PLAIN0_HTML1);
         else
            mdptr = MessageBodyIdentify (vmptr, OPTIONS_READ_PLAIN1_HTML0);
         if (mdptr)
         {
            /* note what we have actually retrieved */
            cdptr->ContentTypeIs = mdptr->ContentTypeIs;
            if (MimeDecProcessPart (mdptr))
            {
               if (mdptr->ContentTypeIs == MIME_CONTENT_TEXT_PLAIN)
               {
                  if (sdptr->MassageMessage)
                     MessageMassage (mdptr->PartContentPtr);
                  cdptr->TextPtr = mdptr->PartContentPtr;
                  cdptr->EditorType = COMPOSE_EDITOR_PLAIN;
               }
               else
               if (mdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
               {
                  HtmlSanitise (mdptr->PartContentPtr);
                  cdptr->TextPtr = mdptr->PartContentPtr;
                  cdptr->EditorType = COMPOSE_EDITOR_HTML;
               }
               else
                  cdptr->TextPtr = "";
            }
            else
               cdptr->TextPtr = "";
            rdptr->CharSetPtr = mdptr->CharsetPtr;
         }
         else
         {
            cdptr->TextPtr = "";
            cdptr->CharSetPtr = uoptr->CharSet;
            if (!vmptr->VmsForeign)
               StatusMessage (FI_LI, 1, "%s", LangFor("mime_type_unknown"));
         }

         if (MsgReply)
         {
            if (vmptr->RfcReplyToPtr)
               cdptr->ToPtr = AddressListAppend (cdptr->ToPtr,
                                                 vmptr->RfcReplyToPtr);
            else
               cdptr->ToPtr = AddressListAppend (cdptr->ToPtr,
                                                 vmptr->MsgFromPtr);
            if (muptr->VmsMailCopyReply) ComposeCopySelf (rdptr);
            cdptr->ToPtr = AddressListMassage (cdptr->ToPtr, FALSE);
            cdptr->AttribPtr = LangFor("compose_a_reply");
         }
         else
         if (MsgReplyAll)
         {
            if (vmptr->RfcReplyToPtr)
               cdptr->ToPtr = AddressListAppend (cdptr->ToPtr,
                                                 vmptr->RfcReplyToPtr);
            else
               cdptr->ToPtr = AddressListAppend (cdptr->ToPtr,
                                                 vmptr->MsgFromPtr);
            if (vmptr->RfcToPtr)
               cdptr->ToPtr = AddressListAppend (cdptr->ToPtr,
                                                 vmptr->MsgToPtr);
            if (vmptr->RfcCcPtr)
               cdptr->CcPtr = AddressListAppend (cdptr->CcPtr,
                                                 vmptr->MsgCcPtr);
            if (muptr->VmsMailCopyReply) ComposeCopySelf (rdptr);
            cdptr->ToPtr = AddressListMassage (cdptr->ToPtr, FALSE);
            cdptr->CcPtr = AddressListMassage (cdptr->CcPtr, FALSE);
            cdptr->AttribPtr = LangFor("compose_a_reply");
         }
         else
         if (MsgForward)
         {
            cdptr->AttribPtr = LangFor("compose_a_forward");
            if (muptr->VmsMailCopyForward) ComposeCopySelf (rdptr);
         }
         if (MsgEditAsNew)
         {
            if (muptr->VmsMailCopySend) ComposeCopySelf (rdptr);
         }

         cdptr->SubjPtr = vmptr->MsgSubjPtr;
         cdptr->FromPtr = vmptr->MsgFromPtr;
      }
      else
      if (muptr->VmsMailCopySend)
         ComposeCopySelf (rdptr);

      StatusMessage (FI_LI, 0, "%s %s",
                     LangFor("compose"), LangFor("page_opened"));
   }
   else
   if (rdptr->MessageId &&
       !strcmp (rdptr->FolderName, uoptr->FolderDraftItems))
   {
      /**************/
      /* draft item */
      /**************/

      vmptr->MessageId = rdptr->MessageId;
      vmptr->ConfirmHash = rdptr->MessageHash;

      CallMailMessageGet (muptr, NULL);

      if (VMSok (muptr->MailVmsStatus))
         DraftRestore (rdptr);
      else
      {
         if (muptr->MailVmsStatus == MAIL$_OPENIN)
         {
            /* error opening external message file */
            StatusMessage (FI_LI, 1, "MAIL: %s %s",
                           SysGetMsg(MAIL$_OPENIN), vmptr->ExtId);
         }
         else
         {
            if (muptr->MailVmsStatus == SS$_BADCHECKSUM ||
                muptr->MailVmsStatus == MAIL$_NOMOREMSG)
               StatusMessage (FI_LI, 1, "%s", LangFor("message_checksum"));
            else
               StatusMessage (FI_LI, 1, "MAIL: %s.",
                              SysGetMsg(muptr->MailVmsStatus));
         }
      }
   }
   else
   {
      /*****************/
      /* anywhere else */
      /*****************/

      if (muptr->VmsMailCopySend) ComposeCopySelf (rdptr);

      rdptr->FieldFocusPtr = "document.getElementById(\'compose_to\').focus();";

      StatusMessage (FI_LI, 0, "%s %s",
                     LangFor("compose"), LangFor("page_opened"));
   }

   ComposePage (rdptr);

   return (TRUE);
}

/*****************************************************************************/
/*
Checks whether the 'to' address is VMS format and if it is compares the
supplied address (that from the selected contact list item) against the 'self'
address (generated in Internet format).  If it is it returns a pointer to the
VMS username.  If not just returns the original (contact list) address.  Used
for non-JavaScript contact list additions to ensure a copy-self is in VMS
format if the 'to' address is (i.e. a VMS mail message).
*/

char* ComposeCheckForSelf
(
REQUEST_DATA *rdptr,
char *AddrPtr
)
{
   COMPOSE_DATA  *cdptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("ComposeCheckForSelf()"); 

   cdptr = &rdptr->ComposeData;

   if (AddressIsVms (cdptr->ToPtr) && !strcmp (AddrPtr, cdptr->SelfPtr))
      return (rdptr->UserName);
   else
      return (AddrPtr);
}

/*****************************************************************************/
/*
Add the 'self' address to the CC list unless the To address is VMS format in
which case use the request username as the VMS Mail address.  Used for
non-JavaScript copy-self if the 'to' address is in VMS format.
*/

void ComposeCopySelf (REQUEST_DATA *rdptr)

{
   COMPOSE_DATA  *cdptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("ComposeCopySelf()"); 

   cdptr = &rdptr->ComposeData;

   if (cdptr->CopiedSelf) return;
   cdptr->CopiedSelf = TRUE;

   if (AddressIsVms (cdptr->ToPtr))
      cdptr->CcPtr = AddressListAppend (cdptr->CcPtr, rdptr->UserName);
   else
      cdptr->CcPtr = AddressListAppend (cdptr->CcPtr, cdptr->SelfPtr);
}

/*****************************************************************************/
/*
Display the message composition (edit) page.
*/

void ComposePage (REQUEST_DATA *rdptr)

{
   int  len,
        AddressTextAreaRows,
        AttachFileCount,
        AttribFwdLength,
        AttribReplyLength,
        ButtonWidth,
        ContactsRows,
        RowSpan,
        SignatureLength;
   char  *cptr, *sptr, *zptr,
         *AddrPtr,
         *AttribPtr,
         *AttribFwdPtr,
         *AttribReplyPtr,
         *ContactList,
         *NamePtr,
         *SignaturePtr;
   COMPOSE_DATA  *cdptr;
   USER_OPTIONS  *uoptr;
   REQUEST_DATA  *sdptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("ComposePage()"); 

   muptr = &VmsMailUser;

   uoptr = &rdptr->UserOptions;
   sdptr = (REQUEST_DATA*)rdptr->StateDataPtr;
   vmptr = &muptr->VmsMailMsg;
   cdptr = &rdptr->ComposeData;

   if (muptr->SignaturePtr)
   {
      /* loaded due to [signature] button */
      SignaturePtr = muptr->SignaturePtr;
      SignatureLength = muptr->SignatureLength;
   }
   else
   {
      SignaturePtr = NULL;
      SignatureLength = 0;
      CGIVARNULL (muptr->SignaturePtr, "FORM_COMPOSE_SIGNATURE");
      if (muptr->SignaturePtr)
         muptr->SignatureLength = strlen(muptr->SignaturePtr);
      else
         OptionsSigFileLoad (rdptr);
   }

   /* if not otherwise set then focus on the message body */
   if (!rdptr->FieldFocusPtr)
      rdptr->FieldFocusPtr = "document.getElementById(\'compose_text\').focus();";

   /* get any contacts list */
   CGIVAR (cptr, "FORM_CONTACT_LIST");
   ContactsLoad (rdptr, cptr, NULL);
   ContactList = TrnLnm (SOY_CONTACT_LIST, NULL, 0);

   cdptr->ToPtr = AddressListMassage (cdptr->ToPtr, FALSE);
   cdptr->CcPtr = AddressListMassage (cdptr->CcPtr, FALSE);
   cdptr->BccPtr = AddressListMassage (cdptr->BccPtr, FALSE);

   if (!cdptr->FromPtr) cdptr->FromPtr = "";
   if (!cdptr->ToPtr) cdptr->ToPtr = "";
   if (!cdptr->CcPtr) cdptr->CcPtr = "";
   if (!cdptr->BccPtr) cdptr->BccPtr = "";
   if (!cdptr->ReplyToPtr) cdptr->ReplyToPtr = "";
   if (!cdptr->SubjPtr) cdptr->SubjPtr = "";
   if (!cdptr->TextPtr) cdptr->TextPtr = "";

   /* get a hash of the current message header (in case it's changed) */
   cdptr->DraftHash = DraftHash (cdptr);

   if (AttribPtr = cdptr->AttribPtr)
   {
      /* do not infinitely increase reply- or forward- tos */
      AttribReplyPtr = LangFor("compose_a_reply");
      AttribReplyLength = strlen(AttribReplyPtr);
      if (strsame (cdptr->SubjPtr, AttribReplyPtr, AttribReplyLength) &&
          strsame (AttribPtr, AttribReplyPtr, AttribReplyLength))
         AttribPtr = "";
      else
      {
         AttribFwdPtr = LangFor("compose_a_forward");
         AttribFwdLength = strlen(AttribFwdPtr);
         if (strsame (cdptr->SubjPtr, AttribFwdPtr, AttribFwdLength) &&
             strsame (AttribPtr, AttribFwdPtr, AttribFwdLength))
            AttribPtr = "";
      }
   }
   else
      AttribPtr = "";

   if (cdptr->CcPtr[0]) cdptr->ShowCcAddr = TRUE;
   if (cdptr->BccPtr[0]) cdptr->ShowBccAddr = TRUE;
   if (cdptr->ReplyToPtr[0]) cdptr->ShowReplyToAddr = TRUE;

   if (AddressIsRfc(cdptr->ToPtr) ||
       AddressIsRfc(cdptr->CcPtr) ||
       AddressIsRfc(cdptr->BccPtr) ||
       AddressIsRfc(cdptr->ReplyToPtr))
      cdptr->LooksLikeRfc = TRUE;
   else
      cdptr->LooksLikeRfc = FALSE;

   RowSpan = 2;
   if (cdptr->ShowCcAddr) RowSpan++;
   if (cdptr->ShowBccAddr) RowSpan++;
   if (cdptr->ShowReplyToAddr) RowSpan++;

   ContactsRows = 3;
   if (cdptr->ShowCcAddr) ContactsRows += 2;
   if (cdptr->ShowBccAddr) ContactsRows += 2;
   if (cdptr->ShowReplyToAddr) ContactsRows += 2;

   if (!cdptr->EditorType)
   {
      /* not explicitly requesting one way or the other */
      if (SoyMailConfig.HtmlEditorLoad)
      {
         if (cdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
            cdptr->EditorType = COMPOSE_EDITOR_HTML;
         else
         if (uoptr->ComposeHTML)
            cdptr->EditorType = COMPOSE_EDITOR_HTML;
         else
            cdptr->EditorType = COMPOSE_EDITOR_PLAIN;
      }
      else
         cdptr->EditorType = COMPOSE_EDITOR_PLAIN;
   }
   /* obviously x 2 */
   if (!SoyMailConfig.HtmlEditorLoad) cdptr->EditorType = COMPOSE_EDITOR_PLAIN;
   if (!rdptr->ClientHasJavaScript) cdptr->EditorType = COMPOSE_EDITOR_PLAIN;

   cptr = LangFor("compose_button_width");
   ButtonWidth = atoi(cptr);
   if (ButtonWidth <= 0 || ButtonWidth > 12) ButtonWidth = 6;
   /* monospace fonts need more width */
   if (uoptr->FontIsMono) ButtonWidth += 2;
   /* allow for the width of "HTML" in the change editor button */
   if (SoyMailConfig.HtmlEditorLoad && rdptr->ClientHasJavaScript)
      ButtonWidth++;

   if (cdptr->EditRows != 20 && cdptr->EditRows != 30 &&
       cdptr->EditRows != 40 && cdptr->EditRows != 60)
      cdptr->EditRows = 20; 
   if (cdptr->EditCols != 72 && cdptr->EditCols != 130)
      cdptr->EditCols = cdptr->WrapAt = 72; 
   if (cdptr->WrapAt && cdptr->WrapAt != cdptr->EditCols)
      cdptr->WrapAt = cdptr->EditCols;

   /* Opera seems to require an extra row - scroll bar one whole row! */
   AddressTextAreaRows = 2;
   if (rdptr->UserAgent == SOY_USER_AGENT_OPERA) AddressTextAreaRows++;

   /**************/
   /* begin page */
   /**************/

   rdptr->CharSetPtr = cdptr->CharSetPtr;
   if (!rdptr->CharSetPtr || !rdptr->CharSetPtr[0])
      rdptr->CharSetPtr = SOY_DEFAULT_CHARSET;

   MainMenuPageBegin (rdptr, PAGE_COMPOSE);
   MainMenuBar (rdptr);
   StatusInfoPanel (rdptr);

   /******************/
   /* message header */
   /******************/

   fputs ("<!-- message header begin -->\n", stdout);

   if (rdptr->ClientHasJavaScript)
   {
      fprintf (stdout,
"<SCRIPT LANGUAGE=\"JavaScript\">\n\
\
function confirmSend () {\n\
   var send = true;\n\
   if (%d) send = confirm(\"%s %s\");\n\
   if (send && document.getElementById(\'compose_to\').value == \"\") \
send = confirm(\"%s %s\");\n\
   if (send && document.getElementById(\'compose_subj\').value == \"\") \
send = confirm(\"%s %s\");\n\
   if (send && document.getElementById(\'compose_text\').value == \"\") \
send = confirm(\"%s %s\");\n\
   return send;\n\
}\n\
\
function confirmSendClose () {\n\
   if(!window.opener) return false;\n\
   return confirmSend();\n\
}\n\
\
function contactListResize() {\n\
   h1 = document.getElementById(\'contact_id_1\').scrollHeight;\n\
   h2 = document.getElementById(\'contact_id_2\').scrollHeight;\n\
   h3 = document.getElementById(\'contact_id_3\').scrollHeight;\n\
   if (h2 > h1) h1 = h2;\n\
   h1 -= h3;\n\
   document.getElementById(\'contact_addr\').style.height = h1;\n\
}\n\
\
function contactNone() {\n\
   refreshStatus (\"status_info\", \
\"<SPAN CLASS=\'css_warning\'>%s</SPAN>\");\n\
   return false;\n\
}\n\
\
function contactNonInternet(addr,field) {\n\
   refreshStatus (\"status_info\", \
\"<SPAN CLASS=\'css_warning\'>\" + field + \": \" + \
addr.replace(/&/g,\"&amp;\").replace(/</g,\"&lt;\").replace(/>/g,\"&gt;\") + \
\"</SPAN>\");\n\
   return false;\n\
}\n\
\
function contactReportAdd(descr,addr) {\n\
   refreshStatus (\"status_info\", descr + \" %s: \" + \
addr.replace(/&/g,\"&amp;\").replace(/</g,\"&lt;\").replace(/>/g,\"&gt;\"));\n\
   return false;\n\
}\n\
\
function contactAdd (which,descr) {\n\
   var trow = document.getElementById(which+\"_tr\");\n\
   var waddr = document.getElementById(which);\n\
   if (which.indexOf(\"compose_reply_to\") != -1) waddr.value = \"\";\n\
   var caddr = document.getElementById(\"contact_addr\");\n\
   if (caddr.selectedIndex == -1) {\n\
      if (waddr.value == \"\") {\n\
         if (which.indexOf(\"compose_to\") != -1) return contactNone();\n\
         if (trow) { \
if (trow.style.display) \
trow.style.display = \"\"; \
else \
trow.style.display = \"none\"; \
}\n\
         contactListResize();\n\
         return false;\n\
      }\n\
      return contactNone();\n\
   }\n\
   var addr = caddr.options[caddr.selectedIndex].value;\n\
   var toval = document.getElementById(\"compose_to\").value;\n\
   if (caddr.selectedIndex == 0 && toval.length > 0 && \
toval.indexOf(\"@\") == -1) return true;\n\
   if (which.indexOf(\"compose_bcc\") != -1 && addr.indexOf(\"@\") == -1) \
return contactNonInternet(addr,\"%s\");\n\
   if (which.indexOf(\"compose_reply_to\") != -1 && addr.indexOf(\"@\") == -1) \
return contactNonInternet(addr,\"%s\");\n\
   caddr.selectedIndex = -1;\n\
   if (waddr.value.indexOf(addr) >= 0) return false;\n\
   if (waddr.value != \"\") waddr.value += \"\\n\";\n\
   waddr.value += addr;\n\
   if (trow && trow.style.display) {\n\
      trow.style.display = \"\";\n\
      contactListResize();\n\
   }\n\
   return contactReportAdd(descr,addr);\n\
}\n",
               cdptr->AlreadySent,
               LangFor("compose_already_sent"), LangFor("sure?"),
               LangFor("compose_no_recipient"), LangFor("sure?"),
               LangFor("compose_no_subject"), LangFor("sure?"),
               LangFor("compose_no_text"), LangFor("sure?"),
               LangFor("compose_add_none"),
               LangFor("compose_add"),
               LangFor("compose_bcc_vms_address"),
               LangFor("compose_reply_to_vms_address"));

      /*********************/
      /* split output here */
      /*********************/

/** Seems to avoid (from the README.TXT) ...
A problem resulting in "unexpected output from the *printf() family
of functions" in VMS V7.3-1 and V7.3-2 that are fixed in VMSV731_ACRTL-V0400
and VMSV732_ACRTL-V0100.
**/

      fprintf (stdout,
"function composeReset() {\n\
   if (!confirm (\"%s\",\"%s\")) return false;\n\
   ct = document.getElementById(\'compose_text\');\n\
   ct.value = ct.defaultValue;\n\
   return false;\n\
}\n\
\
function composeClear() {\n\
   if (!confirm (\"%s\",\"%s\")) return false;\n\
   document.getElementById(\'compose_text\').value =\"\";\n\
   refreshStatus (\"status_info\",\"%s\");\n\
   return false;\n\
}\n\
\
function composeNew() {\n\
   if (!confirm (\"%s\",\"%s\")) return false;\n\
   if (document.getElementById(\"compose_fwd_count\")) return true;\n\
   document.getElementById(\"compose_to\").value = \"\";\n\
   document.getElementById(\"compose_cc\").value = \"\";\n\
   document.getElementById(\"compose_cc_tr\").style.display = \"none\";\n\
   document.getElementById(\"compose_bcc\").value = \"\";\n\
   document.getElementById(\"compose_bcc_tr\").style.display = \"none\";\n\
   document.getElementById(\"compose_reply_to\").value = \"\";\n\
   document.getElementById(\"compose_reply_to_tr\").style.display = \"none\";\n\
   document.getElementById(\"compose_subj\").value = \"\";\n\
   document.getElementById(\"compose_text\").value = \"\";\n\
   document.getElementById(\"compose_to\").focus();\n\
   contactListResize();\n\
   refreshStatus (\"status_info\",\"%s\");\n\
   return false;\n\
}\n\
\
function signatureAppend() {\n\
   var tid = document.getElementById(\"compose_text\");\n\
   var sid = document.getElementById(\"compose_signature\");\n\
   tid.value = tid.value + \"\\n\" + sid.value;\n\
   refreshStatus (\"status_info\",\"%s\");\n\
   return false;\n\
}\n\
\
function setEditRows() {\n\
   var btn = document.getElementById(\"compose_btn_rows\");\n\
   var ct = document.getElementById(\'compose_text\');\n\
   switch (ct.rows) {\n\
      case 20 : ct.rows = 30; break;\n\
      case 30 : ct.rows = 40; break;\n\
      case 40 : ct.rows = 60; break;\n\
      default : ct.rows = 20;\n\
   }\n\
   document.getElementById(\"compose_edit_rows\").value = ct.rows;\n\
   btn.value = \"   \" + ct.rows;\n\
%s\
   ct.focus();\n\
   return false;\n\
}\n\
\
function setEditCols() {\n\
   var btn = document.getElementById(\"compose_btn_cols\");\n\
   var ct = document.getElementById(\'compose_text\');\n\
   switch (ct.cols) {\n\
      case 72 : ct.cols = 130; break;\n\
      default : ct.cols = 72;\n\
   }\n\
   document.getElementById(\"compose_edit_cols\").value = ct.cols;\n\
   btn.value = ct.cols + \"  \";\n\
   ct.focus();\n\
   return false;\n\
}\n\
\
</SCRIPT>\n",
               LangFor("compose_reset_confirm"), LangFor("sure?"), 
               LangFor("compose_clear_confirm"), LangFor("sure?"), 
               LangFor("compose_clear_done"),
               LangFor("compose_new_confirm"), LangFor("sure?"), 
               LangFor("compose_new_done"),
               LangFor("compose_sign_done"),
/*
It appears as if Mozilla (1.7.13) and Firefox (1.5.0.4) does not correctly
resize a <textarea> on a change of the number of rows.  If you change the
number of columns it does.  Hence this kludge, where it is increased slightly
in width then an timer event later decreased by the same amount.  Voila!
*/
               rdptr->UserAgent == SOY_USER_AGENT_MOZILLA ?
"   ct.cols=ct.cols+1;setTimeout(\"ct=document.getElementById(\'compose_text\');\
ct.cols=ct.cols-1\",1) //Mozilla silliness!\n" : "");
   }

   fprintf (stdout,
"<P CLASS=\"css_p08\">\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"0\" CELLSPACING=\"5\" \
CLASS=\"css_panel_1\" ALIGN=\"center\">\n\
\
<TR><TD STYLE=\"width:1%%;vertical-align:top;text-align:right;\">\n\
\
<TABLE CELLPADDING=\"2\" CELLSPACING=\"0\" ID=\"contact_id_1\">\n\
\
<TR><TD CLASS=\"css_tabhdr\">%s:</TD><TD>\
<TEXTAREA CLASS=\"css_text_edit\" STYLE=\"vertical-align:top;width:40em;\" \
ID=\"compose_to\" NAME=\"compose_to\" \
ROWS=\"%d\" COLS=\"40\" WRAP=\"off\" TABINDEX=\"1\">%s</TEXTAREA>\
</TD></TR>\n",
            LangFor("to"),
            AddressTextAreaRows,
            HTML_ADDR(cdptr->ToPtr));

   if (cdptr->ShowCcAddr ||
       rdptr->ClientHasJavaScript)
      fprintf (stdout,
"<TR ID=\"compose_cc_tr\"%s>\
<TD CLASS=\"css_tabhdr\">%s:</TD><TD>\
<TEXTAREA CLASS=\"css_text_edit\" STYLE=\"width:40em;\" \
ID=\"compose_cc\" NAME=\"compose_cc\" ROWS=\"%d\" COLS=\"40\" \
WRAP=\"off\" TABINDEX=\"2\">%s</TEXTAREA>\
</TD></TR>\n",
            cdptr->ShowCcAddr ? "" : " STYLE=\"display:none;\"",      
            LangFor("cc"),
            AddressTextAreaRows,
            HTML_ADDR(cdptr->CcPtr));

   if (cdptr->ShowBccAddr ||
       rdptr->ClientHasJavaScript)
      fprintf (stdout,
"<TR ID=\"compose_bcc_tr\"%s>\
<TD CLASS=\"css_tabhdr\">%s:</TD><TD>\
<TEXTAREA CLASS=\"css_text_edit\" STYLE=\"width:40em;\" \
ID=\"compose_bcc\" NAME=\"compose_bcc\" ROWS=\"%d\" COLS=\"40\" \
WRAP=\"off\" TABINDEX=\"3\">%s</TEXTAREA>\
</TD></TR>\n",
               cdptr->ShowBccAddr ? "" : " STYLE=\"display:none;\"",      
               LangFor("bcc"),
               AddressTextAreaRows,
               HTML_ADDR(cdptr->BccPtr));

   if (cdptr->ShowReplyToAddr ||
       rdptr->ClientHasJavaScript)
      fprintf (stdout,
"<TR ID=\"compose_reply_to_tr\"%s>\
<TD CLASS=\"css_tabhdr\">%s:</TD><TD>\
<INPUT CLASS=\"css_text_edit\" STYLE=\"width:40em;\" \
ID=\"compose_reply_to\" NAME=\"compose_reply_to\" \
SIZE=\"40\" VALUE=\"%s\" TABINDEX=\"4\">\
</TD></TR>\n",
               cdptr->ShowReplyToAddr ? "" : " STYLE=\"display:none;\"",      
               LangFor("compose_reply_to"),
               HTML_ADDR(cdptr->ReplyToPtr));

   fprintf (stdout,
"<TR><TD CLASS=\"css_tabhdr\">%s:</TD><TD>\
<INPUT CLASS=\"css_text_edit\" STYLE=\"width:40em;\" \
ID=\"compose_subj\" NAME=\"compose_subj\" \
SIZE=\"40\" VALUE=\"%s%s%s\" TABINDEX=\"5\">\
</TD></TR>\n",
            LangFor("subject"),
            AttribPtr, *AttribPtr ? " " : "",
            HTML_TEXT(cdptr->SubjPtr));

   fprintf (stdout,
"</TABLE>\n\
\
</TD><TD STYLE=\"width:1%%;padding-left:10;vertical-align:top;text-align:right;\">\n\
\
<TABLE CELLPADDING=\"1\" CELLSPACING=\"0\" ID=\"contact_id_2\">\n\
<TR><TD STYLE=\"padding-top:3\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\" \
ONCLICK=\"return contactAdd(\'compose_to\',\'%s\')\">\
</TD></TR>\n\
<TR><TD>\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\" \
ONCLICK=\"return contactAdd(\'compose_cc\',\'%s\')\">\
</TD></TR>\n\
<TR><TD>\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\" \
ONCLICK=\"return contactAdd(\'compose_bcc\',\'%s\')\">\
</TD></TR>\n\
<TR><TD>\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\" \
ONCLICK=\"return contactAdd(\'compose_reply_to\',\'%s\')\">\
</TD></TR>\n\
</TABLE>\n\
\
</TD><TD STYLE=\"width:98%%;vertical-align:top;text-align:left;\">\n\
\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"2\" CELLSPACING=\"0\">\n\
<TR><TD>\n",
            ButtonWidth,
            LangFor("compose_to"), LangFor("to"),
            ButtonWidth,
            LangFor("compose_cc"), LangFor("cc"),
            ButtonWidth,
            LangFor("compose_bcc"), LangFor("bcc"),
            ButtonWidth,
            LangFor("compose_reply_to"), LangFor("reply_to"));

   ContactsCompose (rdptr, ContactsRows);

   fprintf (stdout,
"</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
<!-- message header end -->\n\n");

   /****************/
   /* message body */
   /****************/

   if (cdptr->EditorType == COMPOSE_EDITOR_HTML)
   {
      fprintf (stdout,
"<!-- HTML editor load -->\n\
%s\n\
<!-- HTML editor loaded! -->\n\
\n",
               SoyMailConfig.HtmlEditorLoad);
      /* good golly miss molly!  is that all it takes?!! */
   }

   fprintf (stdout,
"<!-- message body begin -->\n\
<P CLASS=\"css_p08\">\
<TABLE CELLPADDING=\"1\" CELLSPACING=\"5\" WIDTH=\"100%%\" \
CLASS=\"css_panel_1\" ALIGN=\"center\">\n\
\
<TR ID=\"compose_edit_tr\">\
<TD WIDTH=\"5%%\" STYLE=\"vertical-align:top;\">\n\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\" \
ONCLICK=\"if (!confirmSend()) return false; else return startWorking();\" \
TABINDEX=\"7\">\n",
               ButtonWidth,
               LangFor("compose_send"));

   if (rdptr->ClientHasJavaScript)
      fprintf (stdout,
"<P CLASS=\"css_p02\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" ID=\"compose_btn_send_close\" VALUE=\"%s\" \
ONCLICK=\"if (!confirmSendClose()) return false; else return startWorking();\">\n\
<SCRIPT LANGUAGE=\"JavaScript\">\
if(!window.opener)document.getElementById(\'compose_btn_send_close\')\
.style.fontStyle=\'italic\';\
</SCRIPT>\n",
               ButtonWidth,
               LangFor("compose_send_close"));

   fprintf (stdout,
"<P CLASS=\"css_p02\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;%s\" \
TYPE=\"submit\" NAME=\"compose_btn\" VALUE=\"%s\"",
               ButtonWidth,
               (!(muptr->SignaturePtr && muptr->SignatureLength)) ?
                   "font-style:italic;" : "",
               LangFor("compose_sign"));

   if (!(muptr->SignaturePtr && muptr->SignatureLength))
      fprintf (stdout, " ONCLICK=\"return false\">\n");
   else
   if (cdptr->EditorType == COMPOSE_EDITOR_HTML)
      fprintf (stdout, ">\n");
   else
      fprintf (stdout, " ONCLICK=\"return signatureAppend();\">\n");

   fprintf (stdout,
"<P CLASS=\"css_p08\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\">\n\
\
<P CLASS=\"css_p02\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\">\n\
\
<P CLASS=\"css_p08\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;%s\" TYPE=\"reset\" \
VALUE=\"%s\" ONCLICK=\"%s\">\n",
               ButtonWidth,
               LangFor("compose_save"),
               ButtonWidth,
               LangFor("compose_restore"),
               ButtonWidth,
               cdptr->EditorType == COMPOSE_EDITOR_HTML ?
                  "font-style:italic;" : "",
               LangFor("compose_reset"),
               cdptr->EditorType == COMPOSE_EDITOR_HTML ?
                  "return false" : "return composeReset()");

   fprintf (stdout,
"<P CLASS=\"css_p02\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\" ",
               ButtonWidth,
               LangFor("compose_clear"));
   if (cdptr->EditorType == COMPOSE_EDITOR_HTML)
      fprintf (stdout, "ONCLICK=\"return confirm(\'%s %s\')\">\n",
               LangFor("compose_clear_confirm"), LangFor("sure?"));
   else
      fprintf (stdout, "ONCLICK=\"return composeClear()\">\n");

   fprintf (stdout,
"<P CLASS=\"css_p02\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\" ",
               ButtonWidth,
               LangFor("compose_new"));
   if (cdptr->EditorType == COMPOSE_EDITOR_HTML)
      fprintf (stdout, "ONCLICK=\"return confirm(\'%s %s\')\">\n",
               LangFor("compose_new_confirm"), LangFor("sure?"));
   else
      fprintf (stdout, "ONCLICK=\"return composeNew()\">\n");

   if (SoyMailConfig.HtmlEditorLoad &&
       rdptr->ClientHasJavaScript)
   {
         fprintf (stdout,
"<P CLASS=\"css_p08\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\">\n\
<P CLASS=\"css_p02\">",
               ButtonWidth,
               cdptr->EditorType == COMPOSE_EDITOR_PLAIN ?
                  LangFor("compose_edit_html") :
                  LangFor("compose_edit_plain"));
   }
   else
      fprintf (stdout, "<P CLASS=\"css_p08\">");

   if (cdptr->EditorType == COMPOSE_EDITOR_PLAIN)
      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\">\n",
               ButtonWidth,
               cdptr->WrapAt ? LangFor("compose_wrap_on") :
                               LangFor("compose_wrap_off"));
   else
      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;font-style:italic;\" \
TYPE=\"submit\" NAME=\"compose_btn\" VALUE=\"%s\" ONCLICK=\"return false\">\n",
               ButtonWidth,
               LangFor("compose_wrap_off"));

   fprintf (stdout,
"<INPUT TYPE=\"hidden\" ID=\"compose_edit_rows\" \
NAME=\"compose_edit_rows\" VALUE=\"%d\">\n\
<INPUT TYPE=\"hidden\" ID=\"compose_edit_cols\" \
NAME=\"compose_edit_cols\" VALUE=\"%d\">\n\
<INPUT TYPE=\"hidden\" ID=\"compose_wrap_at\" \
NAME=\"compose_wrap_at\" VALUE=\"%d\">\n\
<P CLASS=\"css_p02\">\
\
<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"width:%1.2fem;border-right-width:0px;\" \
TYPE=\"submit\" ID=\"compose_btn_rows\" NAME=\"compose_btn_rows\" \
VALUE=\"   %d\" ONCLICK=\"%s\">\
\
<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"width:%1.2fem;border-left-width:0px;border-right-width:0px;\" \
TYPE=\"submit\" NAME=\"compose_btn_rows_by_cols\" \
VALUE=\" x \" ONCLICK=\"return false\">\
\
<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"width:%1.2fem;border-left-width:0px;\" \
TYPE=\"submit\" ID=\"compose_btn_cols\" NAME=\"compose_btn_cols\" \
VALUE=\"%d  \" ONCLICK=\"%s\">\n",
            cdptr->EditRows,
            cdptr->EditCols,
            cdptr->WrapAt,
            (float)ButtonWidth / 2.5,
            cdptr->EditRows,
            cdptr->EditorType == COMPOSE_EDITOR_PLAIN ?
               "return setEditRows()" : "return true",
            (float)ButtonWidth / 5.0,
            (float)ButtonWidth / 2.5,
            cdptr->EditCols,
            cdptr->EditorType == COMPOSE_EDITOR_PLAIN ?
               "return setEditCols()" : "return true");

   fprintf (stdout,
"</TD>\
<TD WIDTH=\"5%%\" \
STYLE=\"height:%dem;padding-left:1em;vertical-align:top;\">\n\
\
<TEXTAREA CLASS=\"css_text_edit\" \
ID=\"compose_text\" NAME=\"compose_text\" \
ROWS=\"%d\" COLS=\"%d\" WRAP=\"%s\" TABINDEX=\"6\">",
            /* plus five to maintain the height when using FCKeditor */
            cdptr->EditorType == COMPOSE_EDITOR_PLAIN ?
               cdptr->EditRows : cdptr->EditRows + 5,
            cdptr->EditRows, cdptr->EditCols,
            cdptr->WrapAt > 0 ? "virtual" : "off");

   if (cdptr->AttribPtr && cdptr->MsgReply)
   {
      cptr = LangFor ("compose_quote");
      fprintf (stdout, cptr,
               HTML_ESCAPE(vmptr->MsgDatePtr),
               HTML_ESCAPE(cdptr->FromPtr));
      cptr = HTML_ESCAPE(cdptr->TextPtr);
      if (cdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
      {
         fputs ("<BR>\n", stdout);
         fputs (cptr, stdout);
      }
      else
      {
         fputs ("\n", stdout);
         while (*cptr)
         {
            sptr = cptr;
            while (*cptr && *cptr != '\n') cptr++;
            if (*cptr) cptr++;
            fprintf (stdout, "%s%*.*s",
                     uoptr->QuoteReply, cptr-sptr, cptr-sptr, sptr);
         }
      }
   }
   else
   if (cdptr->AttribPtr && cdptr->MsgForward)
   {
      if (cdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
      {
         fprintf (stdout,
"<BR><BR>--------------- %s ---------------\n\
<BR>%s: %s\n\
<BR>%s: %s\n\
<BR>%s: %s\n\
<BR>%s: %s\n\
<BR><BR>\n",
                  LangFor("compose_original"),
                  LangFor("subject"), HTML_TEXT(vmptr->MsgSubjPtr),
                  LangFor("to"), HTML_ADDR(vmptr->MsgToPtr),
                  LangFor("from"), HTML_ADDR(vmptr->MsgFromPtr),
                  LangFor("date"), HTML_ADDR(vmptr->MsgDatePtr));
      }
      else
      {
         fprintf (stdout,
"\n\n%s--------------- %s ---------------\n\
%s%s: %s\n\
%s%s: %s\n\
%s%s: %s\n\
%s%s: %s\n\
%s\n",
                  uoptr->QuoteForward,
                  LangFor("compose_original"),
                  uoptr->QuoteForward,
                  LangFor("subject"), HTML_TEXT(vmptr->MsgSubjPtr),
                  uoptr->QuoteForward,
                  LangFor("to"), HTML_ADDR(vmptr->MsgToPtr),
                  uoptr->QuoteForward,
                  LangFor("from"), HTML_ADDR(vmptr->MsgFromPtr),
                  uoptr->QuoteForward,
                  LangFor("date"), HTML_TEXT(vmptr->MsgDatePtr),
                  uoptr->QuoteForward);
      }

      if (cdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
      {
         cptr = cdptr->TextPtr;
         fputs (cptr, stdout);
      }
      else
      if (uoptr->QuoteForward[0])
      {
         cptr = HTML_ESCAPE(cdptr->TextPtr);
         while (*cptr)
         {
            sptr = cptr;
            while (*cptr && *cptr != '\n') cptr++;
            if (*cptr) cptr++;
            fprintf (stdout, "%s%*.*s",
                     uoptr->QuoteForward, cptr-sptr, cptr-sptr, sptr);
         }
      }
      else
      {
         cptr = HTML_ESCAPE(cdptr->TextPtr);
         fputs (cptr, stdout);
      }
   }
   else
   if (cdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
      fputs (cdptr->TextPtr, stdout);
   else
      fputs (HTML_ESCAPE(cdptr->TextPtr), stdout);

   if (cdptr->MsgReply) ComposeReplyAttachList (rdptr);

   if (SignaturePtr && SignatureLength)
   {
      if (cdptr->EditorType == COMPOSE_EDITOR_HTML)
      {
         cptr = SignaturePtr;
         while (*cptr)
         {
            sptr = cptr;
            while (*cptr && *cptr != '\n') cptr++;
            if (*cptr) cptr++;
            fprintf (stdout, "<BR>%*.*s", cptr-sptr, cptr-sptr, sptr);
         }
      }
      else
         fprintf (stdout, "\n%s", HTML_TEXT(SignaturePtr));
   }

   fprintf (stdout, "</TEXTAREA>\n</TD><TD WIDTH=\"90%%\"></TD></TR>\n");

   if (cdptr->MsgForward) ComposeForwardAttachList (rdptr);

   fprintf (stdout, "</TABLE>\n");

   if (muptr->SignaturePtr && muptr->SignatureLength)
      fprintf (stdout,
"<INPUT TYPE=\"hidden\" NAME=\"compose_signature\" \
ID=\"compose_signature\" VALUE=\"%s\">\n",
               HTML_ESCAPE(muptr->SignaturePtr));

   fprintf (stdout, "<!-- message body end -->\n\n");

   /***************/
   /* attachments */
   /***************/

   AttachComposeList (rdptr);

   /************/
   /* end page */
   /************/

   MainMenuPageEnd (rdptr);
}

/*****************************************************************************/
/*
Provide each attachment (part with a name or file-name) as "<<name>>" for
inclusion in the body of a replied-to message.
*/

void ComposeReplyAttachList (REQUEST_DATA *rdptr)

{
   char  *cptr, *qptr;
   COMPOSE_DATA  *cdptr;
   MIME_DATA  *mdptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ComposeReplyAttachList()\n");

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;
   cdptr = &rdptr->ComposeData;
   uoptr = &rdptr->UserOptions;

   if (!vmptr->MimeDataPtr) return;

   if (cdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
      fprintf (stdout, "\n<blockquote>\n");
   else
   {
      qptr = uoptr->QuoteReply;
      fprintf (stdout, "\n%s\n", qptr);
   }

   for (mdptr = vmptr->MimeDataPtr; mdptr; mdptr = mdptr->NextMimePtr)
   {
      if (!(mdptr->NamePtr || mdptr->FileNamePtr)) continue;
      if (mdptr->NamePtr)
         cptr = mdptr->NamePtr;
      else
         cptr = mdptr->FileNamePtr;
      if (cdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
         fprintf (stdout, "[[%s]]\n", HTML_ESCAPE(cptr));
      else
         fprintf (stdout, "%s  [[%s]]\n", qptr, cptr);
   }

   if (cdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
      fprintf (stdout, "</blockquote>\n");
}                              

/*****************************************************************************/
/*
Provide each attachment (part with a name or file-name) as a checkbox and
associated name and content-type below the message body text area.  This is
used to selectively [forward] any attachments with a message.
*/

void ComposeForwardAttachList (REQUEST_DATA *rdptr)

{
   BOOL  OriginalForward;
   int  cnt, FwdCount;
   char  *cptr, *nptr, *tptr;
   char  buf [64];
   MIME_DATA  *mdptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ComposeForwardAttachList()\n");

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;

   CGIVARNULL (cptr, "FORM_COMPOSE_FWD_COUNT");
   if (cptr)
   {
      /* this is the compose page at a subsequent request */
      FwdCount = atoi(cptr);
      if (FwdCount < 0 || FwdCount > 100) ErrorExit (SS$_BUGCHECK, FI_LI);
      /* if no attachments to forward */
      if (!FwdCount) return;
      OriginalForward = FALSE;
   }
   else
   {
      /* this is the compose page at the original [forward] request */
      FwdCount = 0;
      for (mdptr = vmptr->MimeDataPtr; mdptr; mdptr = mdptr->NextMimePtr)
         if (mdptr->NamePtr || mdptr->FileNamePtr) FwdCount++;
      /* if no attachments to forward */
      if (!FwdCount) return;
      mdptr = vmptr->MimeDataPtr;
      OriginalForward = TRUE;
   }

   fprintf (stdout,
"\n<!-- forwarded attachments begin -->\n\
<TR><TD></TD><TD STYLE=\"padding-left:1.5em;padding-bottom:0.8em;\">\n\
<INPUT TYPE=\"hidden\" NAME=\"compose_fwd_count\" \
ID=\"compose_fwd_count\" VALUE=\"%d\">\n",
            FwdCount);

   for (cnt = 1; cnt <= FwdCount; cnt++)
   {
      if (OriginalForward)
      {
         /* this is the compose page at the original [forward] request */
         while (mdptr && !(mdptr->NamePtr || mdptr->FileNamePtr))
         {
            mdptr = mdptr->NextMimePtr;
            continue;
         }
         if (!mdptr) break;

         cptr = " CHECKED";
         if (mdptr->NamePtr)
            nptr = HTML_ESCAPE(mdptr->NamePtr);
         else
            nptr = HTML_ESCAPE(mdptr->FileNamePtr);
         tptr = HTML_ESCAPE(mdptr->ContentTypePtr);

         mdptr = mdptr->NextMimePtr;
      }
      else
      {
         /* this is the compose page at a subsequent request */
         sprintf (buf, "FORM_COMPOSE_FWD_%d", cnt);
         CGIVARNULL (cptr, buf);
         if (cptr) cptr = " CHECKED"; else cptr = "";

         sprintf (buf, "FORM_COMPOSE_FWD_NAME_%d", cnt);
         CGIVARNULL (nptr, buf);
         if (!nptr) ErrorExit (SS$_BUGCHECK, FI_LI);

         sprintf (buf, "FORM_COMPOSE_FWD_TYPE_%d", cnt);
         CGIVARNULL (tptr, buf);
         if (!tptr) ErrorExit (SS$_BUGCHECK, FI_LI);

         nptr = HTML_ESCAPE(nptr);
         tptr = HTML_ESCAPE(tptr);
      }

      fprintf (stdout,
"%s<INPUT TYPE=\"checkbox\" NAME=\"compose_fwd_%d\" VALUE=\"1\"%s>\
&nbsp;%s&nbsp;&nbsp;<I>(%s)</I>\n\
<INPUT TYPE=\"hidden\" NAME=\"compose_fwd_name_%d\" VALUE=\"%s\">\n\
<INPUT TYPE=\"hidden\" NAME=\"compose_fwd_type_%d\" VALUE=\"%s\">\n",
               cnt == 1 ? "" : "<BR>",
               cnt, cptr,
               nptr, tptr,
               cnt, nptr,
               cnt, tptr);
   }

   fprintf (stdout, "</TD></TR>\n\
<!-- forwarded attachments end -->\n\n");
}                              

/*****************************************************************************/
/*
Ensure the text of a message is reasonably formatted.
This includes wrapping lines at 72/130/~ characters if not quoted.
The 'escape sequence' of a backslash ('\') plus the (user-option) quote
character suppresses wrapping of that line (the 'escape sequence' is removed).
*/

char* ComposeBodyTextMassage
(
char *TextPtr,
int MaxLineLen,
char *QuoteString
)
{
   int  cnt, lcnt, llen;
   char  *aptr, *cptr, *sptr, *wsptr,
         *NewTextPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ComposeBodyTextMassage() |%s|\n", QuoteString);

   if (MaxLineLen <= 0) MaxLineLen = 72;

   /* first, a quick scan to see if any changes are necessary */
   lcnt = 0;
   cptr = TextPtr;
   while (*cptr)
   {
      if (*QuoteString == *cptr)
         while (*cptr && *cptr != '\n' && *cptr != '\r') cptr++;
      else
      {
         llen = 0;
         while (*cptr && *cptr != '\n' && *cptr != '\r')
         {
            llen++;
            cptr++;
            if (llen <= MaxLineLen) continue;
            lcnt++;
            llen = 0;
         }
      }
      while (*cptr == '\n' || *cptr == '\r') cptr++;
      if (*cptr == '\\' && *(cptr+1) == *QuoteString) lcnt++;
   }
   /* if none required then just return the original message */
   if (!lcnt) return (TextPtr);

   cnt = cptr - TextPtr + (lcnt * strlen(QuoteString)) + 1;
   sptr = NewTextPtr = calloc (cnt, 10);
   if (!sptr) ErrorExit (vaxc$errno, FI_LI); 

   cptr = TextPtr;
   while (*cptr)
   {
      if (*cptr == '\\' && *(cptr+1) == *QuoteString)
      {
         /* remove the 'escape sequence' and do not wrap the line */
         cptr += 2;
         while (*cptr && *cptr != '\n' && *cptr != '\r') *sptr++ = *cptr++;
      }
      else
      if (*cptr == *QuoteString)
      {
         /* 'quoted' - do not wrap the line */
         while (*cptr && *cptr != '\n' && *cptr != '\r') *sptr++ = *cptr++;
      }
      else
      {
         llen = 0;
         while (*cptr && *cptr != '\n' && *cptr != '\r')
         {
            wsptr = NULL;
            while (*cptr && *cptr != '\n' && *cptr != '\r')
            {
               if (*cptr == ' ' || *cptr == '\t')
               {
                  wsptr = sptr;
                  while (*cptr == ' ' || *cptr == '\t')
                  {
                     *sptr++ = *cptr++;
                     llen++;
                  }
                  if (llen >= MaxLineLen)
                  {
                     *(sptr = wsptr) = '\n';
                     sptr++;
                     llen = 0;
                     wsptr = NULL;
                  }
               }
               else
               {
                  *sptr++ = *cptr++;
                  llen++;
                  if (llen >= MaxLineLen && wsptr)
                  {
                     while (*(wsptr+1) == ' ' || *(wsptr+1) == '\t') wsptr++;
                     *wsptr++ = '\n';
                     for (llen = 0; *wsptr; wsptr++) llen++;
                     wsptr = NULL;
                  }
               }
            }
            if (*cptr == '\n')
            {
               *sptr++ = *cptr++;
               llen = 0;
            }
            else
            if (*cptr)
               cptr++;
         }
      }
      while (*cptr == '\n' || *cptr == '\r') *sptr++ = *cptr++;
   }
   *sptr = '\0';

   return (NewTextPtr);
}                              

/*****************************************************************************/
/*
Create a 'self' address for use as the [SELF] contact and as the SMTP/RFC
"From:" header field.
*/

void ComposeSelfAddress (REQUEST_DATA *rdptr)

{
   int  len;
   char  *cptr, *fhptr, *pnptr, *sptr, *zptr;
   char  UserNameLo [64];
   COMPOSE_DATA  *cdptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ComposeSelfAddress()\n");

   muptr = &VmsMailUser;
   cdptr = &rdptr->ComposeData;
   uoptr = &rdptr->UserOptions;

   len = 16;  /* elbow room */
   pnptr = NULL;
   if (uoptr->PersonalName[0])
      len += strlen(pnptr = uoptr->PersonalName);
   else
   if (muptr->VmsMailPersonalName[0])
      len += strlen(pnptr = muptr->VmsMailPersonalName);

   fhptr = NULL;
   /* user option can just be a host/domain name for user@host/domain */
   if (uoptr->SmtpFrom[0]) fhptr = uoptr->SmtpFrom;
   if (!fhptr) fhptr = SoyMailConfig.SmtpFromHost;
   /* if not then try for MX logical names */
   if (!fhptr) fhptr = TrnLnm ("MX_FROM_HOST", NULL, 0);
   if (!fhptr) fhptr = TrnLnm ("MX_ENVELOPE_FROM_HOST", NULL, 0);
   if (!fhptr) fhptr = TrnLnm ("MX_NODE_NAME", NULL, 0);
   /* fall back to the IP host name */
   if (!fhptr) fhptr = LocalHostName;
   if (fhptr) len += strlen(fhptr);

   if (AddressIsRfc(fhptr))
   {
      /* full 'whomever@wherever' specified (usually in the user option) */
      sptr = calloc (1, len);
      if (!sptr) ErrorExit (vaxc$errno, FI_LI);
      if (pnptr)
         sprintf (sptr, "%s <%s>", pnptr, fhptr);
      else
         strcpy (sptr, fhptr);
   }
   else
   {
      /* just the host/domain name */
      zptr = (sptr = UserNameLo) + sizeof(UserNameLo)-1;
      for (cptr = rdptr->UserName;
           *cptr && sptr < zptr;
           *sptr++ = tolower(*cptr++));
      *sptr = '\0';
      len += sptr - UserNameLo;
      sptr = calloc (1, len);
      if (!sptr) ErrorExit (vaxc$errno, FI_LI);
      if (pnptr)
         sprintf (sptr, "%s <%s@%s>",
                  pnptr, UserNameLo, fhptr);
      else
         sprintf (sptr, "%s@%s", UserNameLo, fhptr);
   }

   cdptr->SelfPtr = sptr;

   if (WatchEnabled) WatchThis ("SELF !AZ", cdptr->SelfPtr); 
}                              

/*****************************************************************************/
/*
This monolithic and over-long function creates a MIME compliant RFC822 message.

If there are forwarded attachments and/or soyMAIL attachment files that have
been checked for inclusion a "multipart/mixed" MIME message is generated. 
Each attachment with a "text/.." content-type is quoted-printable encoded in a
separate part.  Attachments with non-text content are provided as base-64
encoded parts.  If there is a message body this is appropriately encoded and
supplied as the first part with a content-type of "text/plain".

If there are no forwarded attachments or files present or checked the message
body is appropriately encoded and made into an in-line MIME message.
*/

BOOL ComposeMime (REQUEST_DATA *rdptr)

{
#define STRINIT(ptr,size)  zptr = (sptr = (char*)ptr) + size;
#define STRCAT(str)        for (cptr=str; *cptr && sptr<zptr; *sptr++=*cptr++);
#define STRNULL            if (sptr<zptr) *sptr = '\0';
#define STROVF             sptr >= zptr

static char  ThisMessageIsInMimeFormat [] =
"This message is in MIME format.  If your mail reader does not understand\r\n\
this format, some or all of the message may not be legible.\r\n";

   int  cnt, len, status,
        AttachFileCount,
        FwdCount,
        DataLength,
        FileSizeBytes;
   char  *cptr, *fnptr, *sptr, *zptr,
         *CgiHttpXForwardedForPtr,
         *CgiRemoteAddrPtr,
         *CgiRemoteHostPtr,
         *CgiRemotePortPtr,
         *CgiServerNamePtr,
         *CharSetPtr,
         *ContentTypePtr,
         *DataPtr,
         *HtmlTextPtr,
         *MessageTextPtr,
         *MimeEncodedPtr,
         *NamePtr,
         *PlainTextPtr,
         *QuotedPrintableTextPtr,
         *TransferEncodingPtr;
   char  AttId [64],
         HeaderBuffer [8192],
         MimeBoundary [64],
         ScratchBuffer [512];
   unsigned long  BinTime [2];
   stat_t  FstatBuffer;
   FILE  *fp;
   COMPOSE_DATA  *cdptr;
   MIME_DATA  *mdptr;
   USER_OPTIONS  *uoptr;
   TEXT_DATA  *tl1ptr, *tl2ptr, *tl3ptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ComposeMime()\n");

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;
   cdptr = &rdptr->ComposeData;
   uoptr = &rdptr->UserOptions;

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   sys$gettim (&BinTime);
   MimeBoundary[0] = '\0';
   CharSetPtr = cdptr->CharSetPtr;
   if (!CharSetPtr || !CharSetPtr[0]) CharSetPtr = SOY_DEFAULT_CHARSET;
   if (Debug) fprintf (stdout, "charset: %s\n", CharSetPtr);

   /****************/
   /* message text */
   /****************/

   if (cdptr->EditorType == COMPOSE_EDITOR_HTML)
      MessageTextPtr = cdptr->TextPtr;
   else
   if (cdptr->WrapAt)
      MessageTextPtr = ComposeBodyTextMassage (cdptr->TextPtr,
                                               cdptr->WrapAt,
                                               uoptr->QuoteReply);
   else
      MessageTextPtr = cdptr->TextPtr;

   HtmlTextPtr = MimeEncodedPtr =
      PlainTextPtr = QuotedPrintableTextPtr = NULL;
   if (cdptr->EditorType == COMPOSE_EDITOR_HTML)
   {
      /* the HTML text body requires 'quoted-printable'ing */
      MimeEncodedPtr = MimeEncQuotedPrintable (MessageTextPtr, -1);
      QuotedPrintableTextPtr = MimeEncodedPtr;
      HtmlTextPtr = MessageTextPtr;
   }
   else
   if (MimeEncNeedsQuotedPrintable (MessageTextPtr))
   {
      /* the plain text body requires 'quoted-printable'ing */
      MimeEncodedPtr = MimeEncQuotedPrintable (MessageTextPtr, -1);
      QuotedPrintableTextPtr = MimeEncodedPtr;
   }
   else
      PlainTextPtr = MimeEncodedPtr = MessageTextPtr;

   tl1ptr = tl2ptr = NULL;
   AttachFileCount = 0;

   /*************************/
   /* check for attachments */
   /*************************/

   /* for forwarded attachments */
   CGIVARNULL (cptr, "FORM_COMPOSE_FWD_COUNT");
   if (cptr)
      FwdCount = atoi(cptr);
   else
      FwdCount = 0;
   if (FwdCount < 0 || FwdCount > 100)
   {
      StatusMessage (FI_LI, 1, "%s", SysGetMsg(SS$_BUGCHECK));
      return (FALSE);
   }

   /* for file-system attachments */
   fnptr = FileSpecSearch (FALSE, muptr->VmsMailFullDirectory, "SOYMAIL_*.ATT");

   /* if either */
   if (FwdCount || fnptr)
   {
      /*******************************/
      /* prepare for multipart/mixed */
      /*******************************/

      /*
         Quite possibly, BUT NOT CERTAINLY, be multipart/mixed!!
         Won't know until we actually process the potential attachment files.
         In the meantime, prepare this first part (the message text) portion
         of a multi-part message.  It will just get overloaded later if no
         attachment files end up getting included (tracked by incrementing
         the counter 'AttachmentCount').
      */
      sprintf (MimeBoundary, "----_=_%08.08X%08.08X",
               BinTime[1], BinTime[0]);

      if (HtmlTextPtr && HtmlTextPtr[0])
      {
         ContentTypePtr = "text/html";
         MimeEncodedPtr = QuotedPrintableTextPtr;
         TransferEncodingPtr = "quoted-printable";
      }
      else
      if (PlainTextPtr && PlainTextPtr[0])
      {
         ContentTypePtr = "text/plain";
         MimeEncodedPtr = PlainTextPtr;
         TransferEncodingPtr = "7bit";
      }
      else
      if (QuotedPrintableTextPtr && QuotedPrintableTextPtr[0])
      {
         ContentTypePtr = "text/plain";
         MimeEncodedPtr = QuotedPrintableTextPtr;
         TransferEncodingPtr = "quoted-printable";
      }
      else
         ContentTypePtr = MimeEncodedPtr = TransferEncodingPtr = NULL;

      /************************************/
      /* "this message is in MIME format" */
      /************************************/

      /* use the 'tl2ptr' expressly so that it can be used to link later */
      len = strlen(ThisMessageIsInMimeFormat);
      TEXT_DATA_CALLOC (tl2ptr, len);
      if (!tl2ptr) ErrorExit (vaxc$errno, FI_LI);

      TEXT_DATA_TEXT(tl2ptr) = TEXT_DATA_SELF(tl2ptr);
      strcpy (TEXT_DATA_TEXT(tl2ptr), ThisMessageIsInMimeFormat);
      TEXT_DATA_LENGTH(tl2ptr) = len;

      /* this IS the first part of the MIME body */
      TEXT_DATA_NEXT (&cdptr->MimeBody) = tl2ptr;

      if (MimeEncodedPtr)
      {
         /****************/
         /* message text */
         /****************/

         if (WatchEnabled)
         {
            WatchThis ("ATTACH (message text)", fnptr); 
            WatchThis ("MIME TYPE !AZ", ContentTypePtr); 
            WatchThis ("MIME TRANSFER !AZ", TransferEncodingPtr); 
         }

         /* first will point to the MIME part header */
         TEXT_DATA_CALLOC (tl1ptr, 512);
         if (!tl1ptr) ErrorExit (vaxc$errno, FI_LI);

         /* link the "this message..." to this new part */
         if (tl2ptr) tl2ptr->NextPtr = tl1ptr;

         /* second will point to the MIME part encoded content */
         TEXT_DATA_CALLOC (tl2ptr, 0);
         if (!tl2ptr) ErrorExit (vaxc$errno, FI_LI);

         TEXT_DATA_TEXT(tl1ptr) = TEXT_DATA_SELF(tl1ptr);
         STRINIT (TEXT_DATA_TEXT(tl1ptr), TEXT_DATA_SIZE(tl1ptr)) 

         STRCAT ("\r\n--")
         STRCAT (MimeBoundary)
         STRCAT ("\r\nContent-Type: ")
         STRCAT (ContentTypePtr)
         STRCAT ("; charset=\"")
         STRCAT (CharSetPtr)
         STRCAT ("\"\r\nContent-Transfer-Encoding: ")
         STRCAT (TransferEncodingPtr)
         STRCAT ("\r\n\r\n")
         STRNULL
         TEXT_DATA_LENGTH (tl1ptr) = sptr - TEXT_DATA_TEXT (tl1ptr);
         if (STROVF) ErrorExit (SS$_BUGCHECK, FI_LI);

         TEXT_DATA_TEXT (tl2ptr) = MimeEncodedPtr;
         TEXT_DATA_LENGTH (tl2ptr) = strlen(TEXT_DATA_TEXT (tl2ptr));

         TEXT_DATA_NEXT (tl1ptr) = tl2ptr;

         if (Debug)
            fprintf (stdout, "|%s|\n|%s|\n",
                     TEXT_DATA_TEXT(tl1ptr), TEXT_DATA_TEXT(tl2ptr));
      }
   }

   /*************************/
   /* forwarded attachments */
   /*************************/

   if (FwdCount)
   {
      CallMailMessageGet (muptr, NULL);
      if (VMSnok (muptr->MailVmsStatus))
      {
         StatusMessage (FI_LI, 1, "MAIL: %s.", SysGetMsg(muptr->MailVmsStatus));
         return (FALSE);
      }
      MessageBodyProcess (vmptr);

      for (cnt = 1; cnt <= FwdCount; cnt++)
      {
         sprintf (ScratchBuffer, "FORM_COMPOSE_FWD_%d", cnt);
         CGIVARNULL (cptr, ScratchBuffer);
         if (!cptr) continue;

         /********************/
         /* checkbox checked */
         /********************/

         AttachFileCount++;

         sprintf (ScratchBuffer, "FORM_COMPOSE_FWD_NAME_%d", cnt);
         CGIVARNULL (cptr, ScratchBuffer);
         if (!cptr) ErrorExit (SS$_BUGCHECK, FI_LI);

         for (mdptr = vmptr->MimeDataPtr; mdptr; mdptr = mdptr->NextMimePtr)
         {
            if (!(mdptr->NamePtr || mdptr->FileNamePtr)) continue;
            if (mdptr->NamePtr)
               NamePtr = mdptr->NamePtr;
            else
               NamePtr = mdptr->FileNamePtr;
            if (!strcmp (NamePtr, cptr)) break;
         }
         if (!mdptr)
         {
            StatusMessage (FI_LI, 1, "%s", SysGetMsg(SS$_BUGCHECK));
            return (FALSE);
         }

         ContentTypePtr = mdptr->ContentTypePtr;
         TransferEncodingPtr = mdptr->TransferEncodingPtr;

         if (!NamePtr || !ContentTypePtr || !TransferEncodingPtr)
         {
            StatusMessage (FI_LI, 1, "%s", SysGetMsg(SS$_BUGCHECK));
            return (FALSE);
         }

         if (WatchEnabled)
         {
            WatchThis ("MIME NAME !AZ", NamePtr); 
            WatchThis ("MIME TYPE !AZ", ContentTypePtr); 
            WatchThis ("MIME TRANSFER !AZ", TransferEncodingPtr); 
         }

         /* first will point to the MIME part header */
         TEXT_DATA_CALLOC (tl1ptr, 512);
         if (!tl1ptr) ErrorExit (vaxc$errno, FI_LI);

         /* if there was a previous part then link it to this new part */
         if (tl2ptr) tl2ptr->NextPtr = tl1ptr;

         /* second will point to the MIME part encoded content */
         TEXT_DATA_CALLOC (tl2ptr, 0);
         if (!tl2ptr) ErrorExit (vaxc$errno, FI_LI);

         TEXT_DATA_TEXT(tl1ptr) = TEXT_DATA_SELF(tl1ptr);
         STRINIT (TEXT_DATA_TEXT(tl1ptr), TEXT_DATA_SIZE(tl1ptr)) 
         STRCAT ("\r\n--")
         STRCAT (MimeBoundary)
         STRCAT ("\r\nContent-Type: ")
         STRCAT (ContentTypePtr)
         STRCAT (";\r\n  name=\"")
         STRCAT (NamePtr)
         STRCAT ("\"\r\nContent-Transfer-Encoding: ")
         STRCAT (TransferEncodingPtr)
         STRCAT ("\r\nContent-Disposition: attachment;\r\n  filename=\"")
         STRCAT (NamePtr)
         STRCAT ("\"\r\n\r\n")
         STRNULL
         TEXT_DATA_LENGTH (tl1ptr) = sptr - TEXT_DATA_TEXT (tl1ptr);
         if (STROVF) ErrorExit (SS$_BUGCHECK, FI_LI);

         /* terminate the MIME content just for these purposes */
         mdptr->PartContentPtr[mdptr->PartContentLength] = '\0';

         TEXT_DATA_TEXT (tl2ptr) = mdptr->PartContentPtr;
         TEXT_DATA_LENGTH (tl2ptr) = mdptr->PartContentLength;

         /* if this is the first part of the MIME body */
         if (!TEXT_DATA_NEXT (&cdptr->MimeBody))
            TEXT_DATA_NEXT (&cdptr->MimeBody) = tl1ptr;

         TEXT_DATA_NEXT (tl1ptr) = tl2ptr;

         if (Debug)
            fprintf (stdout, "%d\n|%s|\n|%s|\n",
                     AttachFileCount, TEXT_DATA_TEXT(tl1ptr),
                     TEXT_DATA_TEXT(tl2ptr));
      }
   }

   /********************/
   /* attachment files */
   /********************/

   while (fnptr)
   {
      /* get the digits from the file name */
      for (cptr = fnptr; *cptr; cptr++);
      while (cptr > fnptr && *cptr != ']') cptr--;
      if (*cptr == ']') cptr++;
      while (*cptr && !isdigit(*cptr)) cptr++;
      zptr = (sptr = AttId) + sizeof(AttId);
      while (isdigit(*cptr) && sptr < zptr) *sptr++ = *cptr++;
      if (sptr > zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';

      if (WatchEnabled) WatchThis ("ATTACH !AZ", fnptr); 

      /* check if the attachment checkbox is currently checked */
      sprintf (ScratchBuffer, "FORM_PART_CBX_%s", AttId);
      CGIVARNULL (sptr, ScratchBuffer);
      /* if not checked just continue */
      if (!sptr)
      {
         fnptr = FileSpecSearch (FALSE, NULL, NULL);
         continue;
      }

      /********************/
      /* checkbox checked */
      /********************/

      AttachFileCount++;

      fp = fopen (fnptr, "r", "ctx=bin");
      if (!fp)
      {
         status = vaxc$errno;
         StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(status));
         return (FALSE);
      }

      if (fstat (fileno(fp), &FstatBuffer))
      {
         status = vaxc$errno;
         StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(status));
         return (FALSE);
      }
      FileSizeBytes = FstatBuffer.st_size;

      DataPtr = calloc (1, FileSizeBytes);
      if (!DataPtr)
      {
         status = vaxc$errno;
         StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(status));
         return (FALSE);
      }

      DataLength = read (fileno(fp), DataPtr, FileSizeBytes);
      if (DataLength <= 0)
      {
         status = vaxc$errno;
         StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(status));
         return (FALSE);
      }
      fclose (fp);

      NamePtr = ContentTypePtr = NULL;

      if (!strncmp (DataPtr, "[soyMAIL-version]", 17))
      {
         sptr = DataPtr + 17;
         while (*sptr) sptr++;
         sptr++;
         if (!strncmp (sptr, "[name]", 6))
         {
            NamePtr = sptr += 6;
            while (*sptr) sptr++;
            sptr++;
            if (!strncmp (sptr, "[type]", 6))
            {
               ContentTypePtr = sptr += 6;
               while (*sptr) sptr++;
               sptr++;
               if (!strncmp (sptr, "[content]", 9)) sptr += 9;
               FileSizeBytes -= sptr - DataPtr;
            }
         }
      }

      if (!NamePtr || !ContentTypePtr)
      {
         StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(SS$_BUGCHECK));
         return (FALSE);
      }

      if (!strncmp (cptr, "text/", 5))
      {
         if (WatchEnabled)
            WatchThis ("MIME QUOTED-PRINTABLE !UL bytes", FileSizeBytes); 
         MimeEncodedPtr = MimeEncQuotedPrintable (sptr, FileSizeBytes);
         if (!MimeEncodedPtr)
         {
            StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(SS$_BUGCHECK));
            return (FALSE);
         }
         TransferEncodingPtr = "quoted-printable";
      }
      else
      {
         if (WatchEnabled) WatchThis ("MIME BASE64 !UL bytes", FileSizeBytes); 
         MimeEncodedPtr = MimeEncBase64 (sptr, FileSizeBytes, "\r\n");
         if (!MimeEncodedPtr)
         {
            StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(SS$_BUGCHECK));
            return (FALSE);
         }
         TransferEncodingPtr = "base64";
      }

      /* free the attachment file content */
      free (DataPtr);

      if (WatchEnabled)
      {
         WatchThis ("MIME NAME !AZ", NamePtr); 
         WatchThis ("MIME TYPE !AZ", ContentTypePtr); 
         WatchThis ("MIME TRANSFER !AZ", TransferEncodingPtr); 
      }

      /* first will point to the MIME part header */
      TEXT_DATA_CALLOC (tl1ptr, 512);
      if (!tl1ptr) ErrorExit (vaxc$errno, FI_LI);

      /* if there was a previous part then link it to this new part */
      if (tl2ptr) tl2ptr->NextPtr = tl1ptr;

      /* second will point to the MIME part encoded content */
      TEXT_DATA_CALLOC (tl2ptr, 0);
      if (!tl2ptr) ErrorExit (vaxc$errno, FI_LI);

      TEXT_DATA_TEXT(tl1ptr) = TEXT_DATA_SELF(tl1ptr);
      STRINIT (TEXT_DATA_TEXT(tl1ptr), TEXT_DATA_SIZE(tl1ptr)) 
      STRCAT ("\r\n--")
      STRCAT (MimeBoundary)
      STRCAT ("\r\nContent-Type: ")
      STRCAT (ContentTypePtr)
      STRCAT (";\r\n  name=\"")
      STRCAT (NamePtr)
      STRCAT ("\"\r\nContent-Transfer-Encoding: ")
      STRCAT (TransferEncodingPtr)
      STRCAT ("\r\nContent-Disposition: attachment;\r\n  filename=\"")
      STRCAT (NamePtr)
      STRCAT ("\"\r\n\r\n")
      STRNULL
      TEXT_DATA_LENGTH (tl1ptr) = sptr - TEXT_DATA_TEXT (tl1ptr);
      if (STROVF) ErrorExit (SS$_BUGCHECK, FI_LI);

      TEXT_DATA_TEXT (tl2ptr) = MimeEncodedPtr;
      TEXT_DATA_LENGTH (tl2ptr) = strlen(TEXT_DATA_TEXT (tl2ptr));

      /* if this is the first part of the MIME body */
      if (!TEXT_DATA_NEXT (&cdptr->MimeBody))
         TEXT_DATA_NEXT (&cdptr->MimeBody) = tl1ptr;

      TEXT_DATA_NEXT (tl1ptr) = tl2ptr;

      if (Debug)
         fprintf (stdout, "%d\n|%s|\n|%s|\n",
                  AttachFileCount, TEXT_DATA_TEXT(tl1ptr),
                  TEXT_DATA_TEXT(tl2ptr));

      fnptr = FileSpecSearch (FALSE, NULL, NULL);
   }

   /******************************/
   /* end of (any) attachment(s) */
   /******************************/

   if (AttachFileCount)
   {
      /* INCLUDED ATTACHMENT(S), add the final trailing MIME boundary */
      TEXT_DATA_CALLOC (tl3ptr, 128);
      if (!tl3ptr) ErrorExit (vaxc$errno, FI_LI);
      TEXT_DATA_TEXT(tl3ptr) = TEXT_DATA_SELF(tl3ptr);
      STRINIT (TEXT_DATA_SELF(tl3ptr), TEXT_DATA_SIZE(tl3ptr)) 
      STRCAT ("\r\n--")
      STRCAT (MimeBoundary)
      STRCAT ("--\r\n")
      STRNULL
      TEXT_DATA_LENGTH (tl3ptr) = sptr - TEXT_DATA_TEXT (tl3ptr);
      if (STROVF) ErrorExit (SS$_BUGCHECK, FI_LI);
      TEXT_DATA_NEXT (tl2ptr) = tl3ptr;
   }
   else
   {
      /* NO ATTACHMENTS, make the text body in-line */
      TEXT_DATA_CALLOC (tl1ptr, 0);
      if (!tl1ptr) ErrorExit (vaxc$errno, FI_LI);
      TEXT_DATA_TEXT (tl1ptr) = MimeEncodedPtr;
      TEXT_DATA_LENGTH (tl1ptr) = strlen(TEXT_DATA_TEXT (tl1ptr));
      /* if this is the only part of the MIME body */
      TEXT_DATA_NEXT (&cdptr->MimeBody) = tl1ptr;
   }

   /***********************************/
   /* create an RFC822 message header */
   /***********************************/

   zptr = (sptr = HeaderBuffer) + sizeof(HeaderBuffer);

   CGIVARNULL (CgiServerNamePtr, "SERVER_NAME");
   if (!CgiServerNamePtr) CgiServerNamePtr = "?";

   CGIVARNULL (CgiRemoteHostPtr, "REMOTE_HOST");
   if (!CgiRemoteHostPtr) CgiRemoteHostPtr = "?";

   CGIVARNULL (CgiRemoteAddrPtr, "REMOTE_ADDR");
   if (!CgiRemoteAddrPtr) CgiRemoteAddrPtr = "?.?.?.?";

   CGIVARNULL (CgiRemotePortPtr, "REMOTE_PORT");
   if (!CgiRemotePortPtr) CgiRemotePortPtr = "?";

   CGIVARNULL (CgiHttpXForwardedForPtr, "HTTP_X_FORWARDED_FOR");

   zptr = (sptr = ScratchBuffer) + sizeof(ScratchBuffer);

   STRCAT ("from ")
   STRCAT (CgiRemoteHostPtr)
   if (strcmp (CgiRemoteHostPtr, CgiRemoteAddrPtr))
   {
      /* resolved host name in REMOTE_HOST */
      STRCAT (" (")
      STRCAT (CgiRemoteAddrPtr)
      if (CgiHttpXForwardedForPtr)
      {
         STRCAT (" forwarded for ")
         STRCAT (CgiHttpXForwardedForPtr)
      }
      STRCAT (")")
   }
   else
   if (CgiHttpXForwardedForPtr)
   {
      /* IP address in REMOTE_HOST */
      STRCAT ("(forwarded for ")
      STRCAT (CgiHttpXForwardedForPtr)
      STRCAT (")")
   }
   STRCAT (" for ")
   STRCAT (rdptr->UserName)
   STRCAT ("@")
   STRCAT (CgiServerNamePtr)
   STRCAT (" by ")
   STRCAT (LocalHostName)
   STRCAT (" (")
   STRCAT (SoftwareId)
   STRCAT (") with HTTP; ")
   STRCAT (MtaRfcBinDate(NULL,NULL))
   STRNULL

   if (STROVF)
   {
      StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
      return (FALSE);
   }

   zptr = (sptr = HeaderBuffer) + sizeof(HeaderBuffer);

   len = InetMailFoldRfc822Header (sptr, zptr-sptr,
                                   "Received:", ScratchBuffer); 
   if (len < 0)
   {
      StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
      return (FALSE);
   }
   sptr += len;

   sptr += sprintf (sptr,
"Message-ID: <%s.%s.%08.08x%08.08x.soymail@%s>\r\nDate: %s\r\n",
                    CgiRemoteAddrPtr, CgiRemotePortPtr,
                    BinTime[1], BinTime[0], LocalHostName,
                    ComposeRfcDate());

   len = InetMailFoldRfc822Header (sptr, zptr-sptr, "Subject:",
                                   MimeEncIsoString(cdptr->SubjPtr,
                                                    CharSetPtr)); 
   if (len < 0)
   {
      StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
      return (FALSE);
   }
   if (len)
      sptr += len;
   else
      STRCAT ("Subject: (none)\r\n")

   len = InetMailFoldRfc822Header (sptr, zptr-sptr, "From:",
                                   MimeEncIsoString(cdptr->FromPtr,
                                                    CharSetPtr)); 
   if (len < 0)
   {
      StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
      return (FALSE);
   }
   sptr += len;

   if (cdptr->ReplyToPtr && cdptr->ReplyToPtr[0])
   {
      len = InetMailFoldRfc822Header (sptr, zptr-sptr, "Reply-to:",
                                      MimeEncIsoString(cdptr->ReplyToPtr,
                                                       CharSetPtr));
      if (len < 0)
      {
         StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
         return (FALSE);
      }
      sptr += len;
   }

   len = InetMailFoldRfc822Header (sptr, zptr-sptr, "To:",
                                   MimeEncIsoString(cdptr->ToPtr,
                                                    CharSetPtr));
   if (len < 0)
   {
      StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
      return (FALSE);
   }
   sptr += len;

   len = InetMailFoldRfc822Header (sptr, zptr-sptr, "CC:",
                                   MimeEncIsoString(cdptr->CcPtr,
                                                    CharSetPtr)); 
   if (len < 0)
   {
      StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
      return (FALSE);
   }
   sptr += len;

   if (cdptr->BccPtr && cdptr->BccPtr[0]) STRCAT ("BCC:\r\n")

   STRCAT ("MIME-Version: 1.0\r\n")

   if (AttachFileCount)
   {
      /* INCLUDED ATTACHMENT(S) so ended up being multipart/mixed */
      STRCAT ("Content-Type: multipart/mixed;\r\n  boundary=\"")
      STRCAT (MimeBoundary)
      STRCAT ("\"\r\n")
   }
   else
   if (HtmlTextPtr)
   {
      /* NO ATTACHMENT(S) so just inline content */
      STRCAT ("Content-Type: text/html; charset=")
      STRCAT (CharSetPtr)
      STRCAT ("\r\nContent-Transfer-Encoding: quoted-printable\r\n")
   }
   else
   if (QuotedPrintableTextPtr)
   {
      /* ditto */
      STRCAT ("Content-Type: text/plain; charset=")
      STRCAT (CharSetPtr)
      STRCAT ("\r\nContent-Transfer-Encoding: quoted-printable\r\n")
   }
   else
   {
      /* ditto again */
      STRCAT ("Content-Type: text/plain; charset=")
      STRCAT (CharSetPtr)
      STRCAT ("\r\nContent-Transfer-Encoding: 7bit\r\n")
   }

   STRCAT ("X-Mailer: ")
   STRCAT (SoftwareId)
   STRCAT ("\r\n\r\n")
   STRNULL

   if (STROVF)
   {
      StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
      return (FALSE);
   }

   len = sptr - HeaderBuffer;
   sptr = calloc (1, len);
   if (!sptr) ErrorExit (vaxc$errno, FI_LI);
   memcpy (sptr, HeaderBuffer, len+1);

   if (WatchEnabled) WatchThis ("RFC822 HEADER\n!AZ", HeaderBuffer);

   TEXT_DATA_TEXT (&cdptr->MimeHeader) = sptr;
   TEXT_DATA_SIZE (&cdptr->MimeHeader) = len+1;
   TEXT_DATA_LENGTH (&cdptr->MimeHeader) = len;

   return (TRUE);
}                              

/*****************************************************************************/
/*
Send the mail message contained in the 'compose data'.
*/

void ComposeSendMessage (REQUEST_DATA *rdptr)

{
   int  len, status;
   char  *cptr, *sptr, *zptr,
         *AddrListPtr,
         *SentItemsFileNamePtr,
         *SentItemsFromPtr,
         *SmtpServerPtr;
   COMPOSE_DATA  *cdptr;
   USER_OPTIONS  *uoptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ComposeSendMessage()\n");

   cdptr = &rdptr->ComposeData;
   uoptr = &rdptr->UserOptions;

   for (cptr = cdptr->ToPtr; *cptr && isspace(*cptr); cptr++);
   if (!*cptr)
   {
      StatusMessage (FI_LI, 1, "%s", LangFor("compose_no_recipient"));
      return;
   }
   if (!rdptr->ClientHasJavaScript)
   {
      for (cptr = cdptr->SubjPtr; *cptr && isspace(*cptr); cptr++);
      if (!*cptr)
      {
         StatusMessage (FI_LI, 1, "%s", LangFor("compose_no_subject"));
         return;
      }
      for (cptr = cdptr->TextPtr; *cptr && isspace(*cptr); cptr++);
      if (!*cptr)
      {
         StatusMessage (FI_LI, 1, "%s", LangFor("compose_no_text"));
         return;
      }
   }

   /* ensure each of the addresses are comma-delimitted */
   cdptr->ToPtr = AddressListMassage (cdptr->ToPtr, TRUE);
   cdptr->CcPtr = AddressListMassage (cdptr->CcPtr, TRUE);
   cdptr->BccPtr = AddressListMassage (cdptr->BccPtr, TRUE);

   if (!cdptr->FromPtr) cdptr->FromPtr = cdptr->SelfPtr;

   /* if it looks like one or more RFC822 addresses */ 
   if (AddressIsRfc(cdptr->ToPtr) ||
       AddressIsRfc(cdptr->CcPtr) ||
       AddressIsRfc(cdptr->BccPtr) ||
       AddressIsRfc(cdptr->ReplyToPtr))
   {
      /************************/
      /* send via an SMTP MTA */
      /************************/

      if (AddressIsVms(cdptr->ToPtr) ||
          AddressIsVms(cdptr->CcPtr) ||
          AddressIsVms(cdptr->BccPtr) ||
          AddressIsVms(cdptr->ReplyToPtr))
      {
         StatusMessage (FI_LI, 1, "%s", LangFor("compose_either_or"));
         return;
      }

      ComposeMime (rdptr);

      /* create a single list of all the addresses for the MTA SMTP server */
      len = strlen(cdptr->ToPtr) +
            strlen(cdptr->CcPtr) +
            strlen(cdptr->BccPtr) + 8;
      AddrListPtr = calloc (1, len);
      if (!AddrListPtr) ErrorExit (vaxc$errno, FI_LI);
      sprintf (AddrListPtr, "%s%s%s%s%s",
               cdptr->ToPtr,
               cdptr->CcPtr[0] ? "\n" : "", cdptr->CcPtr,
               cdptr->BccPtr[0] ? "\n" : "", cdptr->BccPtr);

      SmtpServerPtr = SoyMailConfig.SmtpServerHost;
      if (!SmtpServerPtr) SmtpServerPtr = "localhost";

      cptr = MtaMsgSmtpToServer (SmtpServerPtr,
                                 SentItemsFromPtr = cdptr->FromPtr,
                                 AddrListPtr,
                                 NULL,
                                 &cdptr->MimeHeader,
                                 NULL,
                                 &cdptr->MimeBody);
      if (cptr)
         StatusMessage (FI_LI, 1, "SMTP: %s", HTML_ESCAPE(cptr));
      else
      if (uoptr->FolderSentItems[0])
      {
         /* message sent OK, place it into the "sent Items" folder */
         SentItemsFileNamePtr = SoyMailTempFileName();
         status = MtaMsgToFile (SentItemsFileNamePtr,
                                SentItemsFromPtr, cdptr->ToPtr,
                                cdptr->CcPtr, cdptr->SubjPtr,
                                &cdptr->MimeHeader,
                                &cdptr->MimeBody);
         if (VMSok (status))
            status = CallMailCopyFile (&VmsMailUser,
                                       "MAIL", uoptr->FolderSentItems,
                                       SentItemsFileNamePtr);
         if (VMSnok (status))
            StatusMessage (FI_LI, 1, "%s: %s.",
                           uoptr->FolderSentItems, SysGetMsg(status));

         while (!remove (SentItemsFileNamePtr));
      }
   }
   else
   {
      /****************************************/
      /* otherwise send via VMS callable mail */
      /****************************************/

      if (cdptr->BccPtr && cdptr->BccPtr[0])
      {
         StatusMessage (FI_LI, 1, "%s: %s", 
                        LangFor("compose_bcc_vms_address"),
                        cdptr->BccPtr);
         return;
      }
      else
      if (cdptr->ReplyToPtr && cdptr->ReplyToPtr[0])
      {
         StatusMessage (FI_LI, 1, "%s: %s",
                        LangFor("compose_reply_to_vms_address"),
                        cdptr->ReplyToPtr);
         return;
      }

      /* ensure that the body of the message is correctly wrapped */
      cdptr->TextPtr = ComposeBodyTextMassage (cdptr->TextPtr,
                                               cdptr->WrapAt,
                                               uoptr->QuoteReply);

      cptr = SendMailMessage (SentItemsFromPtr = rdptr->RemoteUser,
                              cdptr->ToPtr,
                              cdptr->CcPtr,
                              cdptr->SubjPtr,
                              cdptr->TextPtr);
      if (cptr)
         StatusMessage (FI_LI, 1, "MAIL: %s", cptr);
      else
      if (uoptr->FolderSentItems[0])
      {
         /* message sent OK, place it into the "sent Items" folder */
         SentItemsFileNamePtr = SoyMailTempFileName();
         status = SendMailToFile (SentItemsFileNamePtr,
                                  SentItemsFromPtr, cdptr->ToPtr,
                                  cdptr->CcPtr, cdptr->SubjPtr,
                                  cdptr->TextPtr);
         if (VMSok (status))
            status = CallMailCopyFile (&VmsMailUser,
                                       "MAIL", uoptr->FolderSentItems,
                                       SentItemsFileNamePtr);
         if (VMSnok (status))
            StatusMessage (FI_LI, 1, "%s: %s.",
                           uoptr->FolderSentItems, SysGetMsg(status));

         while (!remove (SentItemsFileNamePtr));
      }
   }

   if (!cptr)
   {
      /* if the 'delete all attachments after sending' checkbox is */
      CGIVARNULL (cptr, "FORM_ATTACH_CBX_DELETE");
      if (cptr) AttachDelete (rdptr, NULL);

      StatusMessage (FI_LI, 0, "%s", LangFor("compose_sent_ok"));
      cdptr->AlreadySent = TRUE;
   }
}                              

/*****************************************************************************/
/*
*/

void ComposeSendClose (REQUEST_DATA *rdptr)

{
   int  NewMailCount;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ComposeSendClose()\n");

   CgiLibResponseHeader (200, "text/html",
                         "Script-Control: X-content-encoding-gzip=0\n%s",
                         rdptr->LoginSetCookiePtr);

   fprintf (stdout,
"<SCRIPT LANGUAGE=\"JavaScript\">window.close()</SCRIPT>\n");
}

/*****************************************************************************/
/*
Check for the JavaScript created and set form field CLIENT_TIME and use this to
set the message origination date/time if available (and makes sense). 
Otherwise just fall back to the time/zone VMS system soyMAIL is running on.
*/

char* ComposeRfcDate ()

{
   int  cnt, status;
   int  IntTime [6];
   unsigned long  BinTime [2];
   unsigned short  NumTime [7];
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ComposeRfcDate()\n");

   CGIVARNULL (cptr, "FORM_CLIENT_TIME");
   if (!cptr) return (MtaRfcBinDate(NULL,NULL));

   cnt = sscanf (cptr, "%d:%d:%d:%d:%d:%d:",
                 &IntTime[0], &IntTime[1], &IntTime[2],
                 &IntTime[3], &IntTime[4], &IntTime[5]);
   if (Debug) fprintf (stdout, "sscanf() %d\n", cnt);
   if (cnt != 6) return (MtaRfcBinDate(NULL,NULL));

   NumTime[0] = IntTime[0];
   NumTime[1] = IntTime[1];
   NumTime[2] = IntTime[2];
   NumTime[3] = IntTime[3];
   NumTime[4] = IntTime[4];
   NumTime[5] = IntTime[5];
   NumTime[6] = 0;

   status = lib$cvt_vectim (&NumTime, &BinTime);
   if (Debug) fprintf (stdout, "lib$cvt_vectim() %%X%08.08X\n", status);
   if (VMSnok (status)) return (MtaRfcBinDate(NULL,NULL)); 

   for (sptr = cptr; *sptr; sptr++);
   while (sptr > cptr && *sptr != ':') sptr--;
   if (*sptr != ':') return (MtaRfcBinDate(NULL,NULL)); 
      
   return (MtaRfcBinDate(BinTime,sptr+1)); 
}

/*****************************************************************************/

