/*****************************************************************************/
/*
                                  folder.c

Provide HTML form selection lists and buttons for VMS Mail files/folders.


COPYRIGHT
---------
Copyright (C) 2005-2007 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY
---------------
27-JAN-2007  MGD  specially process PMDF IMAP folder subdirectories
30-DEC-2006  MGD  remove "(none)" folder list options
27-DEC-2006  MGD  FolderButtonListEach() simplify button width calculation
21-OCT-2006  MGD  multiple Mail file [copy][move] destination selectors
                    provide all files (drawers) and associated folders 
20-JUL-2006  MGD  try to make font width allowances for monospace
04-JUL-2006  MGD  FolderButtonListEach() refine (perhaps) button width kludge
09-APR-2006  MGD  FolderOpenSingleFileListEach() support public wildcard
                    folder configuration with MAIL and NEWMAIL prohibition
06-APR-2006  MGD  remove default MAIL and NEWMAIL folders
02-APR-2006  MGD  FolderOpenSelect() modifications for public wildcard access
07-FEB-2006  MGD  folder buttons for left-side of message listing
01-FEB-2005  MGD  initial
*/

/*****************************************************************************/

#ifdef SOYMAIL_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* VMS related header files */
#include <descrip.h>
#include <mailmsgdef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header file */
#include "soymail.h"
#include "attach.h"
#include "callmail.h"
#include "cgilib.h"
#include "compose.h"
#include "folder.h"
#include "mainmenu.h"
#include "newmail.h"
#include "other.h"
#include "message.h"

#define FI_LI __FILE__, __LINE__

/********************/
/* external storage */
/********************/

extern BOOL  Debug,
             WatchEnabled;

extern int  VmsVersion;

extern char  CurrentVmsTimeString[],
             SoftwareId[],
             SoftwareVn[];

extern USER_OPTIONS  UserOptions;
extern REQUEST_DATA  RequestData;
extern VMS_MAIL_USER  VmsMailUser;


/*****************************************************************************/
/*
Reset (empty) the mail file and folder lists.
Called upon certain triggers and periodically.
*/

void FolderListReset (REQUEST_DATA *rdptr)

{
   int  cnt, status;
   char  *cptr, *sptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   muptr = &VmsMailUser;

   if (WatchEnabled) WatchThis ("FolderListReset()");

   if (muptr->MailFolderListPtr)
   {
      free (muptr->MailFolderListPtr);
      muptr->MailFolderListPtr = NULL;
      muptr->MailFolderListSize =
         muptr->MailFolderListLength =
         muptr->MailFolderListCount = 0;
   }

   if (muptr->DestFolderListPtr)
   {
      free (muptr->DestFolderListPtr);
      muptr->DestFolderListPtr = NULL;
      muptr->DestFolderListSize =
         muptr->DestFolderListLength =
         muptr->DestFolderListCount = 0;
   }
}

/*****************************************************************************/
/*
Get the list of mail files and a list of folders in the default and destination
mail file (as applicable).  This is quite messy as only one mail file may be
open at a time so we must dance around the primary mail file and folder list.
*/

void FolderListInit (REQUEST_DATA *rdptr)

{
   int  cnt, len, status,
        VmsMailFileNameLength;
   char  *cptr, *sptr, *zptr;
   char  VmsMailFileName [256];
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;

   if (WatchEnabled) WatchThis ("FolderListInit()");

   if (!uoptr->UseOnlyMAILFile &&
       !muptr->DestFolderListPtr)
   {
      if (!muptr->MailFileListPtr) CallMailFileList (muptr);

      /* buffer original mail file name */
      zptr = (sptr = VmsMailFileName) + sizeof(VmsMailFileName)-1;
      for (cptr = muptr->VmsMailFileName;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';
      VmsMailFileNameLength = muptr->VmsMailFileNameLength;

      /* substitute with the destination mail file */
      zptr = (sptr = muptr->VmsMailFileName) + sizeof(muptr->VmsMailFileName)-1;
      for (cptr = rdptr->DestMailFileName;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';
      muptr->VmsMailFileNameLength = sptr - muptr->VmsMailFileName;

      status = CallMailFileClose (muptr);
      if (VMSnok (status)) return;

      status = CallMailFolderList (muptr);
      if (VMSnok (status))
         /* it's quite possible a specified mail file does not exist (yet) */
         if (status != MAIL$_OPENIN)
            StatusMessage (FI_LI, 1, "MAIL: %s.", SysGetMsg(status));

      /* restore original mail file name */
      zptr = (sptr = muptr->VmsMailFileName) + sizeof(muptr->VmsMailFileName)-1;
      for (cptr = VmsMailFileName; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';
      muptr->VmsMailFileNameLength = VmsMailFileNameLength;

      status = CallMailFileClose (muptr);
      if (VMSnok (status)) return;

      /* copy the data from the current folder list then reset it */
      if (muptr->MailFolderListPtr)
      {
         muptr->DestFolderListPtr = muptr->MailFolderListPtr;
         muptr->DestFolderListSize = muptr->MailFolderListSize;
         muptr->DestFolderListLength = muptr->MailFolderListLength;
         muptr->DestFolderListCount = muptr->MailFolderListCount;
         muptr->MailFolderListPtr = NULL;
         muptr->MailFolderListSize =
            muptr->MailFolderListLength =
            muptr->MailFolderListCount = 0;
      }

      time ((unsigned long*)&rdptr->FolderRefreshSeconds);
   }

   if (!muptr->MailFolderListPtr)
   {
      if (muptr->DestFolderListPtr &&
          !strcmp (rdptr->MailFileName, rdptr->DestMailFileName))
      {
         /* current and destination files same, clone dest folder list */
         muptr->MailFolderListCount = muptr->DestFolderListCount;
         muptr->MailFolderListLength = muptr->DestFolderListLength;
         muptr->MailFolderListSize = muptr->DestFolderListLength + 1;
         muptr->MailFolderListPtr = calloc (1, muptr->MailFolderListSize);
         if (!muptr->MailFolderListPtr) ErrorExit (vaxc$errno, FI_LI);
         memcpy (muptr->MailFolderListPtr,
                 muptr->DestFolderListPtr,
                 muptr->MailFolderListSize);
         muptr->MailFolderListMaxLen = muptr->DestFolderListMaxLen;
      }
      else
      {
         status = CallMailFolderList (muptr);
         if (VMSnok (status))
         {
            if (muptr->MailVmsStatus == MAIL$_OPENIN &&
                muptr->MailFileDeletedCount)
               StatusMessage (FI_LI, -1, LangFor("mail_file_deleted"),
                              rdptr->MailFileName);
            else
               StatusMessage (FI_LI, 1, "MAIL: %s.", SysGetMsg(status));
         }
         time ((unsigned long*)&rdptr->FolderRefreshSeconds);
         muptr->MailFolderListMaxLen = 0;
         muptr->MailFolderListMaxLen = muptr->DestFolderListMaxLen;
      }
   }

   /* find the length of the longest mail file name */
   muptr->MailFolderListMaxLen = 0;
   cptr = muptr->MailFileListPtr;
   for (cnt = muptr->MailFileListCount; cnt; cnt--)
   {
      if (*cptr == '[')
      {
         /* it's a PMDF IMAP subfolder */
         while (*cptr && *cptr != ']') cptr++;
         if (*cptr) cptr++;
      }
      if ((len = strlen(cptr)) > muptr->MailFolderListMaxLen)
         muptr->MailFolderListMaxLen = len;
      while (*cptr) cptr++;
      cptr++;
   }
   muptr->DestFolderListMaxLen = muptr->MailFolderListMaxLen;

   /* integrate with the length of the longest folder name */
   cptr = muptr->MailFolderListPtr;
   for (cnt = muptr->MailFolderListCount; cnt; cnt--)
   {
      if ((len = strlen(cptr)) > muptr->MailFolderListMaxLen)
         muptr->MailFolderListMaxLen = len;
      while (*cptr) cptr++;
      cptr++;
   }

   /* and with the length of the longest destination folder name */
   cptr = muptr->DestFolderListPtr;
   for (cnt = muptr->DestFolderListCount; cnt; cnt--)
   {
      if ((len = strlen(cptr)) > muptr->DestFolderListMaxLen)
         muptr->DestFolderListMaxLen = len;
      while (*cptr) cptr++;
      cptr++;
   }

   if (Debug)
      fprintf (stdout, "mail:%d dest:%d\n",
               muptr->MailFolderListMaxLen, muptr->DestFolderListMaxLen);
}

/*****************************************************************************/
/*
Generate a selection list for the available mail folders and [open] button. 
If private access then include the [9 new] button.
*/

void FolderOpenSelect (REQUEST_DATA *rdptr)

{
   int  status;
   char  newBuf [256];
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderOpenSelect()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;

   fprintf (stdout,
"<!-- folder open begin -->\n\
<INPUT TYPE=\"hidden\" NAME=\"folder_list_change\" \
ID=\"folder_list_change\" VALUE=\"\">\n\
<SELECT CLASS=\"css_select_2\" NAME=\"folder_list_open\"%s \
ONCHANGE=\"document.getElementById(\'folder_list_change\').value=true;\
startWorking();document.getElementById(\'soymail_form\').submit();\">\n",
            muptr->MailFolderListMaxLen > 20 ? " STYLE=\"width:12em;\"" : "");

   FolderOpenListEach (rdptr);

   sprintf (newBuf, "%d %s", rdptr->NewMessages, LangFor("new"));

   fprintf (stdout,
"</SELECT>&nbsp;&nbsp;\n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"main_btn\" VALUE=\"%s\"%s>&nbsp;\n",
            LangFor("open"),
            CLICK_STARTWORKING);

   if (rdptr->PrivateAccess)
      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"main_btn\" VALUE=\"%s\"%s>&nbsp; \n",
               newBuf,
               NewMailInclude(rdptr) ?
                  " ONCLICK=\"refreshNewMail(-1);return false;\"" :
                  CLICK_STARTWORKING);

   fprintf (stdout,
"<!-- folder open end -->\n\n");
}

/*****************************************************************************/
/*
*/

void FolderOpenListEach (REQUEST_DATA *rdptr)

{
   int  cnt;
   char  *cptr, *sptr, *tptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderOpenListEach()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;

   if (rdptr->PublicWildcard ||
       uoptr->UseOnlyMAILFile ||
       muptr->MailFileListCount <= 1)
      FolderOpenSingleFileListEach (rdptr);
   else
      FolderOpenMultiFileListEach (rdptr);
}

/*****************************************************************************/
/*
Only a single MAIL.MAI file or optioned to only show MAIL.MAI.
List each folder.
*/

void FolderOpenSingleFileListEach (REQUEST_DATA *rdptr)

{
   int  fldcnt;
   char  *cptr, *fldptr, *escptr, *selptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderOpenSingleFileListEach()");

   muptr = &VmsMailUser;

   if (!rdptr->NewMessages && !strcmp (rdptr->FolderName, "NEWMAIL"))
      fputs ("<OPTION VALUE=\"NEWMAIL\" STYLE=\"padding-left:0.5em;\" \
SELECTED>NEWMAIL\n", stdout);

   fldptr = muptr->MailFolderListPtr;
   for (fldcnt = muptr->MailFolderListCount; fldcnt; fldcnt--)
   {
      if (rdptr->PublicWildcard &&
          rdptr->PublicNoNewMail &&
          (!strcmp (fldptr, "MAIL") ||
           !strcmp (fldptr, "NEWMAIL")))
      {
         /* next folder in list */
         while (*fldptr) fldptr++;
         fldptr++;
         continue;
      }

      if (!strcmp (fldptr, rdptr->FolderName))
         selptr = " SELECTED";
      else
         selptr = "";
      escptr = HTML_ESCAPE(fldptr);
      fprintf (stdout,
"<OPTION VALUE=\"%s\" STYLE=\"padding-left:0.5em;\"%s>%s\n",
               escptr, selptr, escptr);

      /* next folder in list */
      while (*fldptr) fldptr++;
      fldptr++;
   }
}

/*****************************************************************************/
/*
Multiple mail .MAI files.
List each folder for the currently selected mail .MAI file.
*/

void FolderOpenMultiFileListEach (REQUEST_DATA *rdptr)

{
   int  filcnt, fldcnt;
   char  *cptr, *filptr, *fldptr, *escptr, *selptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderOpenMultiFileListEach()");

   muptr = &VmsMailUser;

   filptr = muptr->MailFileListPtr;
   for (filcnt = muptr->MailFileListCount; filcnt; filcnt--)
   {
      cptr = filptr;
      if (*cptr == '[')
      {
         /* it's a PMDF IMAP subfolder */
         while (*cptr && *cptr != ']') cptr++;
         if (*cptr) cptr++;
      }

      if (!strcmp (filptr, rdptr->MailFileName))
      {
         /* current mail file */
         fprintf (stdout, "<OPTION VALUE=\"%s/MAIL\" SELECTED>%s%s/\n",
                  /* PMDF IMAP mail files are indicated by a double slash */
                  filptr, cptr, cptr == filptr ? "" : "/");
         fldptr = muptr->MailFolderListPtr;
         for (fldcnt = muptr->MailFolderListCount; fldcnt; fldcnt--)
         {
            if (!strcmp (fldptr, rdptr->FolderName))
               selptr = " SELECTED";
            else
               selptr = "";

            escptr = HTML_ESCAPE(fldptr);
            /* Opera 9 doesn't notice the style padding hence the &nbsp; */
            fprintf (stdout,
"<OPTION VALUE=\"%s/%s\" STYLE=\"padding-left:1em;\"%s>&nbsp;&nbsp;/%s\n",
                     filptr, escptr, selptr, escptr);

            /* next folder name in the list */
            while (*fldptr) fldptr++;
            fldptr++;
         }
      }
      else
      {
         /* show an empty folder for each mail file */
         fprintf (stdout, "<OPTION VALUE=\"%s/MAIL\">%s%s/\n",
                  /* PMDF IMAP mail files are indicated by a double slash */
                  filptr, cptr, cptr == filptr ? "" : "/");
      }

      /* next mail file name in the list */
      while (*filptr) filptr++;
      filptr++;
   }
}

/*****************************************************************************/
/*
Differences in the CSS rendering of (only) buttons inside table data (the CSS
lay-up technicalities are well outside any expertise I might have) for Firefox
1.5, Mozilla 1.7, Opera 8.5 and MSIE 6.0 make this guesstimate of the width of
the required buttons absolutely necessary.  I played around with the various
bits of CSS I understood and still got vastly differing behaviours between the
agents and eventually resorted to this nightmare.  Talk about making it up as
you go along!
*/

void FolderButtonListEach (REQUEST_DATA *rdptr)

{
   int  ButtonWidth;
   float  WidthFactor;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderButtonListEach()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;

   ButtonWidth = muptr->MailFolderListMaxLen;

   if (ButtonWidth > 30)
      WidthFactor = 0.70;
   else
   if (ButtonWidth > 20)
      WidthFactor = 0.65;
   else
   if (ButtonWidth > 10)
      WidthFactor = 0.80;
   else
      WidthFactor = 1.10;

   /* MSIE6 seems to render fonts slightly larger than the others tested */
   if (rdptr->UserAgent == SOY_USER_AGENT_MSIE) WidthFactor += 0.1;

   /* who uses monospace?! (at least the one person who reported issues :-) */
   if (uoptr->FontIsMono) WidthFactor += 0.15;

   ButtonWidth = (int)((float)ButtonWidth * WidthFactor);

   if (uoptr->UseOnlyMAILFile || muptr->MailFileListCount <= 1)
      FolderButtonSingleFileListEach (rdptr, ButtonWidth);
   else
      FolderButtonMultiFileListEach (rdptr, ButtonWidth);
}

/*****************************************************************************/
/*
Only a single MAIL.MAI file or optioned to only show MAIL.MAI.
Provide a button for each folder.
*/

void FolderButtonSingleFileListEach
(
REQUEST_DATA *rdptr,
int ButtonWidth
)
{
   int  cnt, fldcnt, idx;
   char  *cptr, *fldptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderButtonSingleFileListEach()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;

   cnt = 0;
   fldptr = muptr->MailFolderListPtr;
   for (fldcnt = muptr->MailFolderListCount; fldcnt; fldcnt--)
   {
      if (cnt && !(cnt % uoptr->FolderButtons)) 
         fprintf (stdout, "<P CLASS=\"css_p02\">");
      cnt++;

      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"width:%dem;text-align:left;padding-left:0.5em;\" \
TYPE=\"submit\" NAME=\"folder_btn\" VALUE=\"%s\">\n",
               ButtonWidth, HTML_ESCAPE(fldptr));

      /* next folder in list */
      while (*fldptr) fldptr++;
      fldptr++;
   }
}

/*****************************************************************************/
/*
Multiple mail .MAI files.
Provide a button for each mail file and each folder for the currently selected
mail .MAI file.
*/

void FolderButtonMultiFileListEach
(
REQUEST_DATA *rdptr,
int ButtonWidth
)
{
   int  cnt, filcnt, fldcnt, idx;
   char  *cptr, *filptr, *fldptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderButtonMultiFileListEach()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;

   /* folder buttons don't contain the mail file, supply using this */
   fprintf (stdout,
"<INPUT TYPE=\"hidden\" NAME=\"folder_btn_mailfile\" VALUE=\"%s\">\n",
            rdptr->MailFileName); 

   cnt = 0;

   filptr = muptr->MailFileListPtr;
   for (filcnt = muptr->MailFileListCount; filcnt; filcnt--)
   {
      if (cnt && !(cnt % uoptr->FolderButtons)) 
         fprintf (stdout, "<P CLASS=\"css_p02\">");
      cnt++;

      /* show a button for each mail file */
      if (rdptr->ClientHasJavaScript)
      {
         cptr = filptr;
         if (*cptr == '[')
         {
            /* it's a PMDF IMAP subfolder */
            while (*cptr && *cptr != ']') cptr++;
            if (*cptr) cptr++;
         }

         fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"width:%dem;text-align:left;padding-left:0.5em;\" \
TYPE=\"submit\" NAME=\"folder_btn\" VALUE=\"%s%s/\" \
ONCLICK=\"this.value=\'%s/\';return true;\">\n",
               /* PMDF IMAP mail files are indicated by a double slash */
               ButtonWidth, cptr, cptr == filptr ? "" : "/", filptr);
      }
      else
         fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"width:%dem;text-align:left;padding-left:0.5em;\" \
TYPE=\"submit\" NAME=\"folder_btn\" VALUE=\"%s%s/\">\n",
               /* PMDF IMAP mail files are indicated by a double slash */
               ButtonWidth, filptr, cptr == filptr ? "" : "/");

      if (!strcmp (filptr, rdptr->MailFileName))
      {
         /* current mail file */
         fldptr = muptr->MailFolderListPtr;
         for (fldcnt = muptr->MailFolderListCount; fldcnt; fldcnt--)
         {
            if (cnt && !(cnt % uoptr->FolderButtons)) 
               fprintf (stdout, "<P CLASS=\"css_p02\">");
            cnt++;

            fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"width:%dem;text-align:left;padding-left:2em;\" \
TYPE=\"submit\" NAME=\"folder_btn\" VALUE=\"/%s\">\n",
                     ButtonWidth, HTML_ESCAPE(fldptr));

            /* next folder name in the list */
            while (*fldptr) fldptr++;
            fldptr++;
         }
      }

      /* next mail file name in the list */
      while (*filptr) filptr++;
      filptr++;
   }
}

/*****************************************************************************/
/*
Provide a selection list for the message action destination folders.
*/

void FolderDestListEach (REQUEST_DATA *rdptr)

{
   int  cnt;
   char  *cptr, *sptr, *tptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderDestListEach()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;

   if (uoptr->UseOnlyMAILFile || muptr->MailFileListCount <= 1)
      FolderDestSingleFileListEach (rdptr);
   else
      FolderDestMultiFileListEach (rdptr);
}

/*****************************************************************************/
/*
List each folder available as a destination for [copy] or [move].
*/

void FolderDestSingleFileListEach (REQUEST_DATA *rdptr)

{
   int  fldcnt, idx;
   char  *cptr, *fldptr, *escptr, *selptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderDestSingleFileListEach()");

   muptr = &VmsMailUser;

   fldptr = muptr->MailFolderListPtr;
   for (fldcnt = muptr->MailFolderListCount; fldcnt; fldcnt--)
   {
      /* it's not any of the default folders */
      if (!strcmp (fldptr, rdptr->DestFolderName))
         selptr = " SELECTED";
      else
         selptr = "";
      escptr = HTML_ESCAPE(fldptr);
      fprintf (stdout,
"<OPTION VALUE=\"%s\" STYLE=\"padding-left:0.5em;\"%s>%s\n",
               escptr, selptr, escptr);
      /* next folder in list */
      while (*fldptr) fldptr++;
      fldptr++;
   }
}

/*****************************************************************************/
/*
List each mail file and possible folder available as a destination for [copy]
or [move].
*/

void FolderDestMultiFileListEach (REQUEST_DATA *rdptr)

{
   int  filcnt, fldcnt;
   char  *cptr, *filptr, *fldptr, *escptr, *selptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderDestMultiFileListEach()");

   muptr = &VmsMailUser;

   filptr = muptr->MailFileListPtr;
   for (filcnt = muptr->MailFileListCount; filcnt; filcnt--)
   {
      cptr = filptr;
      if (*cptr == '[')
      {
         /* it's a PMDF Imap subfolder */
         while (*cptr && *cptr != ']') cptr++;
         if (*cptr) cptr++;
      }

      if (!strcmp (filptr, rdptr->DestMailFileName))
      {
         /* current destination mail file */
         fprintf (stdout, "<OPTION VALUE=\"%s/MAIL\" SELECTED>%s%s/\n",
                  /* PMDF IMAP mail files are indicated by a double slash */
                  filptr, cptr, cptr == filptr ? "" : "/");

         fldptr = muptr->DestFolderListPtr;
         for (fldcnt = muptr->DestFolderListCount; fldcnt; fldcnt--)
         {
            if (!strcmp (fldptr, rdptr->DestFolderName))
               selptr = " SELECTED";
            else
               selptr = "";

            escptr = HTML_ESCAPE(fldptr);
            /* Opera 9 doesn't notice the style padding hence the "&nbsp"; */
            fprintf (stdout,
"<OPTION VALUE=\"%s/%s\" STYLE=\"padding-left:1em;\"%s>&nbsp;&nbsp;/%s\n",
                     filptr, escptr, selptr, escptr);

            /* next folder name in the list */
            while (*fldptr) fldptr++;
            fldptr++;
         }
      }
      else
         fprintf (stdout, "<OPTION VALUE=\"%s/MAIL\">%s%s/\n",
                  /* PMDF IMAP mail files are indicated by a double slash */
                  filptr, cptr, cptr == filptr ? "" : "/");

      /* next mail file name in the list */
      while (*filptr) filptr++;
      filptr++;
   }
}

/*****************************************************************************/
/*
A faked folder select specifically for the help menu bar page.
*/

void FolderOpenHelp (REQUEST_DATA *rdptr)

{
   int  status;
   char  newBuf [256];
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderOpenHelp()");

   fprintf (stdout,
"<!-- folder open help begin -->\n\
<TD NOWRAP ALIGN=\"left\" VALIGN=\"middle\">\n\
<SELECT CLASS=\"css_select_2\" NAME=\"folder_list_help\">\n\
<OPTION VALUE=\"%s\" SELECTED>&nbsp;%s&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;\n\
</SELECT>&nbsp;&nbsp;\n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"main_btn\" VALUE=\"%s\">&nbsp; \n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"main_btn\" VALUE=\"0 %s\">&nbsp; \n\
</TD>\n\
<!-- folder open help end -->\n\n",
            LangFor("help"), LangFor("help"),
            LangFor("open"),
            LangFor("new"));
}

/*****************************************************************************/
/*
Generate an appropriate folder description (i.e. if it's not the default mail
file then include it in the description) and return a pointer to dynamic
storage.

The contortions detecting the '/' in both the mail file name and the folder
name is to allow PMDF IMAP folders supplied in the format
"[01M8RV66VU1I95NYS4]MAIL//Whatever" to be processed as well as where the mail
file and folder names each contain discrete components.
*/

char* FolderDescribe
(
char *MailFileName,
char *FolderName
)
{
   char  *cptr, *sptr, *zptr;
   char  buf [256];

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderDescribe()");

   zptr = (sptr = buf) + sizeof(buf)-1;
   if (MailFileName)
   {
      if (strcmp (MailFileName, "MAIL"))
      {
         /* not the default MAIL.MAI */
         cptr = MailFileName;
         if (*cptr == '[')
         {
            /* it's a PMDF Imap subfolder */
            while (*cptr && *cptr != ']') cptr++;
            if (*cptr) cptr++;
         }
         if (cptr == MailFileName)
         {
            while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
            for (cptr = " / "; *cptr && sptr < zptr; *sptr++ = *cptr++);
         }
         else
         {
            while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
            /* PMDF IMAP mail files are indicated by a double slash */
            for (cptr = " // "; *cptr && sptr < zptr; *sptr++ = *cptr++);
         }
      }
   }
   if (FolderName)
   {
      for (cptr = FolderName; *cptr && *cptr != '/'; cptr++);
      if (*cptr) cptr++;
      if (!*cptr)
         for (cptr = FolderName; *cptr && sptr < zptr; *sptr++ = *cptr++);
   }
   *sptr = '\0';
   cptr = calloc (1, sptr-buf+1);
   strcpy (cptr, buf);
   return (cptr);
}

/*****************************************************************************/

